# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
def factorial(n) :
   r"""
{xrst_begin docstring_example}

Docstring Example
#################

Discussion
**********
This example demonstrates using a python docstring to document a function.
see :ref:`indent_example@Python Docstring`
for an example of an indented docstring.

This Example File
*****************
{xrst_literal}

{xrst_end docstring_example}
   """
   if n == 1 :
      return 1
   return n * factorial(n-1)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin spell_cmd user}
{xrst_spell
   abcd
   cpp
   index index
}

Spell Command
#############

Syntax
******
- ``\{xrst_spell`` *word_1* ...  *word_n* ``}``
- ``\{xrst_spell_off}``
- ``\{xrst_spell_on}``

The lines containing the ``{`` ( ``}`` ) character
must have nothing but white space before ( after )  it.

spell_off
*********
By default xrst does spell checking.
You can turn spell checking off using this command.

spell_on
********
If spell checking has been turned off,
you can turn it back on using this command.

spell
*****
You can specify special words to include as correct spelling for
this page using this command.

Words
*****
Each word, that is checked for spelling, is a sequence of letters.
Upper case letters start a new word (even when preceded by a letter).

Special Words
*************
In the syntax above, the special word list is

| *word_1* ... *word_n*

These words are considered correct spelling even though
they are not in the dictionary.
In the syntax above the special words are all in one line.
They could be on different lines which helps when displaying
the difference between  versions of the corresponding file.
Latex commands should not be in the special word list because
words that have a backslash directly before them
are not include in spell checking.

project_dictionary
******************
The list of words in the
:ref:`config_file@project_dictionary`
are considered correct spellings for all pages.
If multiple people are working on a project using different spell checkers,
the words that are correct in one spell checker and not another should
be included in the project_dictionary.

page_name
*********
For each of the following commands in a page, the words in *page_name*
are considered correct spelling for that page:

| |tab| ``\{xrst_begin``        *page_name* *group_name* ``}``
| |tab| ``\{xrst_begin_parent`` *page_name* *group_name* ``}``
| |tab| ``:ref:`` \` ...  *page_name*\ ``-name`` ... `
| |tab| ``:ref:`` \` ...  *page_name*\ ``-title`` ... `

Note that *group_name* can be empty which corresponds to the default group;
see :ref:`begin_cmd-name` .

Capital Letters
***************
The case of the first letter does not matter when checking spelling;
e.g., if ``abcd`` is *word_1* then ``Abcd`` will be considered a valid word.
Each capital letter starts a new word; e.g., `CamelCase` is considered to
be the two words 'camel' and 'case'.
Single letter words are always correct and not included in the
special word list; e.g., the word list entry ``CppAD`` is the same as ``Cpp``.

Double Words
************
It is considered an error to have only white space between two occurrences
of the same word. You can make an exception for this by entering
the same word twice (next to each other) in the special word list.

Double words errors occur in the output the user sees.
for example, the input:
::

   `python package index <https://pypi.org/>`_ index.

results in the double word 'index index' in the output the user sees; i.e.,
the following output:
`python package index <https://pypi.org/>`_ index.

Example
*******
:ref:`spell_example-name`

{xrst_end spell_cmd}
"""
# ---------------------------------------------------------------------------
import sys
import re
import xrst
#
# pattern
pattern = dict()
pattern['spell']     = re.compile(
   r'[^\\]{xrst_spell([^_a-z][^}]*)}' +
   r'([ \t]*@xrst_line ([0-9]+)@)?'
)
pattern['word_error'] = re.compile( r'[^A-Za-z \t\n]' )
#
# pattern
# global pattern values used by spell command
pattern['template']  = xrst.pattern['template_begin']
pattern['dir']       = xrst.pattern['dir']
pattern['toc']       = xrst.pattern['toc']
pattern['code']      = xrst.pattern['code']
pattern['line']      = xrst.pattern['line']
pattern['literal']   = xrst.pattern['literal']
#
# pattern
# local pattern values only used by spell command
pattern['directive']  = re.compile( r'\n[ ]*[.][.][ ]+[a-z-]+::' )
pattern['http']       = re.compile( r'(https|http)://[A-Za-z0-9_/.#-]*' )
pattern['ref_name_1'] = re.compile( r':ref:`([^\n<`]+)-name`' )
pattern['ref_name_2'] = re.compile( r':ref:`[^\n<`]*<([^\n>`]+)-name *>`' )
pattern['ref_1']      = re.compile( r':ref:`[^\n<`]+`' )
pattern['ref_2']      = re.compile( r':ref:`([^\n<`]+)<[^\n>`]+>`' )
pattern['url_1']      = re.compile( r'`<[^\n>`]+>`_' )
pattern['url_2']      = re.compile( r'`([^\n<`]+)<[^\n>`]+>`_' )
#
# pattern['spell_on'], pattern['spell_off']
lin = r'[ \t]*@xrst_line ([0-9]+)@'
pattern['spell_on']  = re.compile( r'\n[ \t]*\{xrst_spell_on}' + lin )
pattern['spell_off'] = re.compile( r'\n[ \t]*\{xrst_spell_off}' + lin )
#
# pattern['word']
# The first choice is for line numbers which are not in original file.
# The second is characters that are not letters, white space, or backslash.
# These character separate double words so they are not an error.
# The third is for the actual words (plus a possible backlash at start).
pattern['word']  = re.compile(
   r'(@xrst_line [0-9]+@|[^A-Za-z\s\\]+|\\?[A-Za-z][a-z]+)'
)
#
# pattern['page_name_word']
# The pattern for a word in the page name.
pattern['page_name_word'] = re.compile( r'[A-Za-z][a-z]+' )
# -----------------------------------------------------------------------------
# {xrst_begin spell_cmd_dev dev}
# {xrst_comment_ch #}
#
# Process the spell command for a page
# ####################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# tmp_dir
# *******
# The file :ref:`replace_spell@spell.toml`
# is written in the *tmp_dir* directory by the one page at a time
# by this function call.
#
# data_in
# *******
# is the data for this page before the spell commands are removed.
#
# page_file
# *********
# is the name of the file where the begin command for this page appears.
# This is used for error reporting and for creating spell.toml.
#
# page_name
# *********
# is the name of the page that this data is in. This is only used
# for error reporting and spell.toml.
#
# begin_line
# **********
# is the line number in *page_file* where the begin command for this page
# appears. This is only used for spell.toml.
#
# ignore_commands
# ***************
# If this is true (false), :ref:`run_xrst@ignore_spell_commands`
# was (was not) on the command line.
#
# print_warning
# *************
# if this is false, do not print the spelling warnings. Otherwise,
# a spelling warning is reported for each word (and double word) that is not
# in the spell_checker dictionary or the special word list. A word is two or
# more letter characters. If a word is directly preceded by a backslash,
# it is ignored (so that latex commands do not generate warnings).
#
# spell_checker
# *************
# Is a spell checking object used for error checking; see
# :ref:`get_spell_checker-name`.
#
# data_out
# ********
# is the data for this page after the spell command (if it exists)
# is removed.
#
# spell_warning
# *************
# is true if a spelling error occurred and a warning was printed.
# Otherwise, it is false.
#
# unknown_word_set
# ****************
# is a lower case version of the set of words that are not in the
# spell_checker dictionary or the special word list.
# It does not include double word errors.
#
# {xrst_end spell_cmd_dev}
# BEGIN_DEF
def spell_command(
   tmp_dir,
   data_in,
   page_file,
   page_name,
   begin_line,
   ignore_commands,
   print_warning,
   spell_checker
) :
   assert type(tmp_dir) == str
   assert type(data_in) == str
   assert type(page_file) == str
   assert type(page_name) == str
   assert type(begin_line) == int
   assert type(print_warning) == bool
   # END_DEF
   #
   # first_spell_warning
   first_spell_warning = True
   #
   # unknown_word_set
   unknown_word_set = set()
   #
   # m_spell
   m_spell       = pattern['spell'].search(data_in)
   #
   # data_out
   data_out = data_in
   #
   # special_list, data_out
   special_list = list()
   if m_spell != None :
      #
      # check for multiple spell commands in one page
      m_error  = pattern['spell'].search(data_in, m_spell.end() )
      if m_error :
         msg  = 'There are two spell xrst commands in this page'
         xrst.system_exit(
            msg,
            file_name=page_file,
            page_name=page_name,
            m_obj=m_error,
            data=data_in
         )
      #
      # word_list
      word_list = m_spell.group(1)
      word_list = pattern['line'].sub('', word_list)
      m_error   = pattern['word_error'].search(word_list)
      if m_error :
         m_line = pattern['line'].search( data_in[m_spell.start() :] )
         line   = int( m_line.group(1) )
         line  += word_list[: m_error.start() ].count('\n')
         ch         = word_list[m_error.start()]
         ascii_code = ord(ch)
         msg  = 'The word list in spell command contains a character\n'
         msg += 'that is not a letter or white space.\n'
         msg += f'ascii code = {ascii_code}, character = {ch}.'
         #
         # The spell command cannot be inside a template file so we can use
         # line number to report the error.
         xrst.system_exit(
            msg,
            file_name=page_file,
            page_name=page_name,
            line = line,
         )
      #
      # special_list
      for m_obj in pattern['word'].finditer( word_list ) :
         word_lower = m_obj.group(0).lower()
         if not word_lower.startswith('@xrst_line') :
            special_list.append( word_lower )
      #
      # remove spell command
      start    = m_spell.start()
      end      = m_spell.end()
      data_out = data_in[: start+1] + data_in[end :]
   #
   # special_list
   if ignore_commands :
      special_list = list()
   #
   # special_used, double_used
   special_used   = dict()
   double_used    = dict()
   previous_word  = ''
   for word_lower in special_list :
      if word_lower == previous_word :
         double_used[ word_lower ] = False
      else :
         special_used[ word_lower ] = False
      #
      previous_word = word_lower
   #
   # data_tmp
   # version of data_in with certain commands removed
   data_tmp = data_out
   #
   # data_tmp
   # remove \< and \>, pattern matching below does not have to worry about it
   data_tmp = data_tmp.replace( r'\<', ' ')
   data_tmp = data_tmp.replace( r'\>', ' ')
   #
   # page_name_word
   page_name_word = list()
   for m_word in pattern['page_name_word'].finditer(page_name) :
      word_lower = m_word.group(0).lower()
      page_name_word.append( word_lower )
   for m_tmp in pattern['ref_name_1'].finditer(data_tmp) :
      page_name_tmp = m_tmp.group(1)
      for m_word in pattern['page_name_word'].finditer(page_name_tmp) :
         word_lower = m_word.group(0).lower()
         page_name_word.append( word_lower )
   for m_tmp in pattern['ref_name_2'].finditer(data_tmp) :
      page_name_tmp = m_tmp.group(1)
      for m_word in pattern['page_name_word'].finditer(page_name_tmp) :
         word_lower = m_word.group(0).lower()
         page_name_word.append( word_lower )
   #
   # m_off
   m_off = pattern['spell_off'].search(data_tmp)
   while m_off :
      #
      # m_on
      m_on = pattern['spell_on'].search(data_tmp, m_off.end() )
      #
      #
      off = m_off.start()
      #
      # on
      if m_on == None :
         on = len( data_tmp )
         m_off = pattern['spell_off'].search(data_tmp, off)
         if m_off :
            msg  = 'There are two spell_off commands '
            msg += 'without a spell_on command between them'
            xrst.system_exit(
               msg,
               file_name=page_file,
               page_name=page_name,
               m_obj=m_off,
               data=data_tmp
            )
      else :
         on = m_on.end()
      #
      # data_tmp
      data_tmp = data_tmp[: off ] + data_tmp[on : ]
      #
      # m_off
      m_off = pattern['spell_off'].search(data_tmp, off)
   #
   m_on = pattern['spell_on'].search(data_tmp)
   if m_on :
      msg  = 'There is a spell_on command '
      msg += 'without a spell_off command before it'
      xrst.system_exit(
         msg,
         file_name=page_file,
         page_name=page_name,
         m_obj=m_on,
         data=data_tmp
      )
   #
   # data_tmp
   # This is for detection of double word errors because these words are
   # automatically in the special word list.
   m_ref = pattern['ref_name_1'].search(data_tmp)
   while m_ref != None :
      page_name_tmp  = m_ref.group(1)
      page_name_words = ' '
      for m_word in pattern['page_name_word'].finditer(page_name_tmp) :
         word_lower       = m_word.group(0).lower()
         if page_name_words.find( f' {word_lower} ' ) < 0 :
            page_name_words += word_lower + ' '
      before   = data_tmp[: m_ref.start() ]
      after    = data_tmp[m_ref.end() :]
      data_tmp = before + ' ' + page_name_words + ' ' + after
      m_ref = pattern['ref_name_1'].search(data_tmp)
   #
   # data_tmp
   # commands with file names as arugments
   # Use @ character to avoid mistaken double word errors
   data_tmp = pattern['dir'].sub('@', data_tmp)
   data_tmp = pattern['literal'].sub('@', data_tmp)
   data_tmp = pattern['toc'].sub('@', data_tmp)
   data_tmp = pattern['http'].sub('@', data_tmp)
   data_tmp = pattern['directive'].sub('@', data_tmp)
   #
   # Must keep template file name error reporting
   # data_tmp = pattern['template'].sub('@', data_tmp)
   #
   # command with page names and headings as arguments
   data_tmp = pattern['ref_1'].sub('@', data_tmp)
   data_tmp = pattern['ref_2'].sub(r'\1', data_tmp)
   data_tmp = pattern['code'].sub('@', data_tmp)
   #
   # commands with external urls as arguments
   data_tmp = pattern['url_1'].sub('@', data_tmp)
   data_tmp = pattern['url_2'].sub(r'\1', data_tmp)
   #
   # any left over xrst commands
   data_tmp = re.sub( r'{xrst_comment_ch' , '@', data_tmp)
   #
   # previous_word
   previous_word = ''
   #
   # replace_word_list
   replace_word_list = list()
   #
   # m_obj
   for m_obj in pattern['word'].finditer( data_tmp ) :
      #
      # word, word_lower
      word       = m_obj.group(0)
      word_lower = word.lower()
      #
      if not word.startswith('@xrst_line') and word[0].isalpha()  :
         #
         # known
         # check if word is in dictionary or word_lower is special
         known =  spell_checker.known( word ) or word_lower in page_name_word
         if not known :
            #
            page_line, template_file, template_line = \
               xrst.file_line(m_obj, data_tmp)
            #
            # replace_word_list
            if not word_lower in replace_word_list :
               replace_word_list.append( word_lower )
            #
            # unknown_word_set
            if not word_lower in special_used :
               unknown_word_set.add( word_lower )
            #
            if not word_lower in special_used and print_warning :
               #
               # first_spell_warning
               if first_spell_warning :
                  msg  = '\nwarning: page_file = ' + page_file
                  msg += ', page_name = ' + page_name + '\n'
                  sys.stderr.write(msg)
                  first_spell_warning = False
               #
               # msg
               msg  = 'spelling = ' + word
               suggest = spell_checker.suggest(word)
               if suggest != None :
                  msg += f', suggest = {suggest}'
               msg += f', page_line = {page_line}'
               if template_file != None :
                  msg += f', template_file = {template_file}'
                  msg += f', template_line = {template_line}'
               msg += '\n'
               #
               sys.stderr.write(msg)
            #
            # special_used
            if word_lower in special_used :
               special_used[word_lower] = True
         #
         double_word = word_lower == previous_word.lower()
         if double_word :
            # This is the second use of word with only white space between
            #
            # replace_word_list
            if not word_lower in replace_word_list :
               replace_word_list.append( word_lower )
               replace_word_list.append( word_lower )
            else :
               index = replace_word_list.index( word_lower )
               assert 0 < index
               if index + 1 < len(replace_word_list) :
                  if replace_word_list[index + 1] != word_lower :
                     replace_word_list.insert(index, word_lower )
               else :
                  replace_word_list.append(word_lower)
            #
            if not word_lower in double_used and print_warning :
               # word is not in list of special double words
               #
               # first_spell_warning
               if first_spell_warning :
                  msg  = 'warning: file = ' + page_file
                  msg += ', page = ' + page_name + '\n'
                  sys.stderr.write(msg)
                  first_spell_warning = False
               #
               # line_number
               m_tmp = pattern['line'].search(data_tmp, m_obj.end() )
               assert m_tmp
               line_number = m_tmp.group(1)
               msg  = f'double word error: "{previous_word} {word}"'
               msg += ', line ' + line_number + '\n'
               sys.stderr.write(msg)
            #
            # double_used
            if word_lower in double_used :
               double_used[word_lower]  = True
      if not word.startswith('@xrst_line') :
         # previous_word
         # This captures when there are non space characters between words
         previous_word = word
   #
   # check for words that were not used
   for word_lower in special_used :
      ok = special_used[word_lower] or word_lower in double_used
      if print_warning and not ok :
         if first_spell_warning :
            msg  = '\nwarning: file = ' + page_file
            msg += ', page = ' + page_name + '\n'
            sys.stderr.write(msg)
            first_spell_warning = False
         msg = 'spelling word "' + word_lower + '" not needed\n'
         sys.stderr.write(msg)
   for word_lower in double_used :
      if not double_used[word_lower] and print_warning :
         if first_spell_warning :
            msg  = '\nwarning: file = ' + page_file
            msg += ', page = ' + page_name + '\n'
            sys.stderr.write(msg)
            first_spell_warning = False
         msg  = 'double word "' + word_lower + ' ' + word_lower
         msg += '" not needed\n'
         sys.stderr.write(msg)
   #
   # start_spell, end_spell
   if m_spell :
      m_line = pattern['line'].search(data_in, m_spell.start() )
      start_spell = int( m_line.group(1) )
      end_spell   = int( m_spell.group(3) )
   else :
      start_spell = 0
      end_spell   = 0
   #
   # begin_line
   # 2DO, remove this check (may not be valid in the future)
   m_line     = pattern['line'].search(data_in)
   check = int( m_line.group(1) )
   assert begin_line == check
   #
   # file_data
   file_data  = f'[ "{page_file}"."{page_name}" ]\n'
   file_data += f'begin_line = {begin_line}\n'
   file_data += f'start_spell = {start_spell}\n'
   file_data += f'end_spell = {end_spell}\n'
   if len(replace_word_list) == 0 :
      file_data += f'replace = []\n\n'
   else :
      file_data += f'replace = [\n'
      for word_lower in replace_word_list :
         file_data += f'   "{word_lower}",\n'
      file_data += ']\n\n'
   #
   # spell.toml
   file_obj   = open(f'{tmp_dir}/spell.toml', 'a')
   file_obj.write(file_data)
   file_obj.close()
   #
   # data_out
   data_out = pattern['spell_off'].sub( '', data_out)
   data_out = pattern['spell_on'].sub( '', data_out)
   #
   #
   spell_warning = not first_spell_warning
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   assert type(spell_warning) == bool
   assert type(unknown_word_set) == set
   return data_out, spell_warning, unknown_word_set
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin run_xrst user}
{xrst_spell
   dev
   furo
   github
   grep
   jax
   pdf
   rtd
   thet
   txt
   wil
}

Extract RST Files And Run Sphinx
################################

Syntax
******

| ``xrst`` \\
| |tab| [ ``--version`` ] \\
| |tab| [ ``--local_toc`` ] \\
| |tab| [ ``--page_source`` ] \\
| |tab| [ ``--external_links`` ] \\
| |tab| [ ``--replace_spell_commands`` ] \\
| |tab| [ ``--ignore_spell_commands`` ] \\
| |tab| [ ``--suppress_spell_warnings`` ] \\
| |tab| [ ``--continue_with_warnings`` ] \\
| |tab| [ ``--rst_line_numbers`` ] \\
| |tab| [ ``--rst_only`` ] \\
| |tab| [ ``--index_page_name`` *index_page_name* ] \\
| |tab| [ ``--config_file``     *config_file* ] \\
| |tab| [ ``--html_theme``      *html_theme* ] \\
| |tab| [ ``--target``          *target* ]  \\
| |tab| [ ``--number_jobs``     *number_jobs* ] \\
| |tab| [ ``--group_list``      *group_name_1* *group_name_2* ... ] \\
| |tab| [ ``--rename_group``    *old_group_name* *new_group_name* ] \\

#. The brackets around each of lines above indicate that the line is optional.
#. The lines above can appear in any order.
#. Text in code font; e.g. ``--target`` is explicit; i.e.,
   must appear exactly as above.
#. Text in italic font; .e.g, *target* is implicit; i.e.,
   it gets replaced by the user's choice.
#. It may be helpful to remove the :ref:`config_file@directory@html_directory`
   before running the command below.
   This will check for error messages that are not repeated due
   to caching the results of previous builds.

{xrst_comment --------------------------------------------------------------- }

version
*******
If ``--version`` is present on the command line,
the version of xrst is printed and none of the other arguments matter.
{xrst_comment --------------------------------------------------------------- }

local_toc
*********
If this option is present on the command line,
a table of contents for the Headings in the current page
is included at the top of every page.
The page name and page title are not in this table of contents.

Some :ref:`html themes<run_xrst@html_theme>` include this information
on a side bar; e.g., ``furo`` and ``sphinx_book_theme`` .
{xrst_comment --------------------------------------------------------------- }

page_source
***********
If this option is present and *target* is ``html`` ,
a link to the xrst source code is included at the top of each page.
Some :ref:`html themes<run_xrst@html_theme>` include this link; e.g.,
``sphinx_rtd_theme`` .

If this option is present and *target* is ``tex`` ,
the xrst source code file is reported at the beginning of each page.

external_links
**************
If this option is present, the external links are checked.
The ones that are broken or redirects are reported.
Broken links are considered errors and redirects are warnings.

replace_spell_commands
**********************
If this option is present on the command line, the source code
:ref:`spell commands<spell_cmd-name>` are replaced in such a way that the
there will be no spelling warnings during future processing by xrst.
If this option is present,
none of the output files are created; e.g., the \*.rst and \*.html files.

#. This is useful when there are no spelling warnings before a change
   to the :ref:`config_file@project_dictionary` or when there is an update
   to the :ref:`config_file@spell_package` .
#. This is also useful when there are no spelling warnings and you want
   to sort the words in all the spelling commands.

See also :ref:`config_file@project_dictionary` .

ignore_spell_commands
*********************
If this option is present on the command line, the
:ref:`spell commands<spell_cmd-name>` are ignored and
none of the output files are created; e.g., the \*.rst and \*.html files.

#. This can be used to find words that should be added to the
   :ref:`config_file@project_dictionary` ; i.e.,
   words in error and with a high page count .
#. If you remove a word from the project dictionary, this can be used
   to check how many pages use that word.

If you change the project dictionary consider using
:ref:`run_xrst@replace_spell_commands` .

suppress_spell_warnings
***********************
If this option is present on the command line, none of the spelling warnings
will be generated.
This is useful when there are no spelling warnings with one spelling package
and you are temporarily using a different version of the package
or a different package altogether.
{xrst_comment --------------------------------------------------------------- }

continue_with_warnings
**********************
If this option is (is not) present on the command line,
the program will not exit (will exit) with an error when warnings are
generated.
{xrst_comment --------------------------------------------------------------- }

rst_line_numbers
****************
Normally sphinx error and warning messages are reported using line numbers
in the xrst source code files.
If this option is present, these messages are reported
using the line numbers in the RST files created by xrst.
In addition the :ref:`run_xrst@page_source` links to the rst files,
instead of the xrst source files.
This may be helpful if you have an error or warning for a sphinx command
and it does not make sense using xrst source code line numbers.
It is also helpful for determining if an incorrect line number is due to
sphinx or xrst.
{xrst_comment --------------------------------------------------------------- }

rst_only
********
Normally, after extraction the RST files,
xrst automatically runs sphinx to produce the target output (html or tex).
If this option is present, sphinx is not run.
Only the rst files, and their corresponding sources,
are generated; i.e.,

| |tab| :ref:`config_file@directory@rst_directory`/\*.rst
| |tab| *rst_directory*\ /_sources/\*.txt

This may be useful when creating rst files for uses else where; e.g.,
for use with `Read the Docs <https://docs.readthedocs.io>`_
(see :ref:`.readthedocs.yaml-name` for a better way to use Read the Docs.)
The sphinx commands are printed after xrst finishes and can be executed
by hand.
This may be useful if there is a problem during these commands.
{xrst_comment --------------------------------------------------------------- }

index_page_name
***************
This option has no effect when *target* is ``tex`` .
If *target* is ``html``,
the file ``index.html`` in the
:ref:`config_file@directory@html_directory` will be a redirect
to the page specified by *index_page_name* .
If this option is not present, ``index.html`` wil be a redirect
to the root of the documentation tree.
{xrst_comment --------------------------------------------------------------- }

config_file
***********
The command line argument *config_file* specifies the location of the
:ref:`config_file-name` for this project.
This can be an absolute path or
relative to the directory where :ref:`xrst<run_xrst-name>` is run.

xrst.toml
=========
If *config_file* is not present on the command line,
the default value ``xrst.toml`` is used for *config_file* .
{xrst_comment --------------------------------------------------------------- }

html_theme
**********
This the html_theme_ that is used.
The default value for *html_theme* is ``furo`` .

.. _html_theme: https://sphinx-themes.org/

Theme Choices
=============
The following is a list of some themes that work well with the
default settings in :ref:`config_file@html_theme_options` .
If you have a theme together with html_theme_options
that work well with xrst,
please post an issue on github so that it can be added to the list below.

{xrst_spell_off}
.. csv-table:: Sphinx Themes
   :header: name,  local_toc

   sphinx_rtd_theme,     yes
   furo,                 no
   sphinx_book_theme,    no
   pydata_sphinx_theme,  no
   piccolo_theme,        no
{xrst_spell_on}

sphinx_rtd_theme
================
The sphinx_rtd theme builds faster than some of the other themes,
so it is suggested to use it for testing (with the ``--local_toc`` option).
A special modification is made to this theme when *target* is html,
so that it displays wider than its normal limit.
This modification may be removed in the future.
{xrst_comment --------------------------------------------------------------- }

target
******
The command line argument *target* must be ``html`` or ``tex``.
It specifies the type of type output you plan to generate using sphinx.
Note thet :ref:`config_file@directory@html_directory` and
:ref:`config_file@directory@tex_directory` will determine the location
of the corresponding output files.
The default value for *target* is ``html`` .

tex
===
If you choose this target, xrst will create the file
*project_name*\ ``.tex`` in the :ref:`config_file@directory@tex_directory` .
There are two reasons to build this file.
One is to create the file *project_name*\ ``.pdf``
which is a pdf version of the documentation.
The other is to test for errors in the latex sections of the documentation.
(MathJax displays latex errors in red, but one has to check
every page that has latex to find all the errors this way.)
Once you have built *project_name*\ ``.tex``, the following command
executed in :ref:`config_file@directory@project_directory`
will accomplish both purposes:

   make -C *tex_directory* *project_name*\ ``.pdf``

#. The :ref:`config_file@project_name` is specified in the configuration file.
#. The resulting output file will be *project*\ ``.pdf`` in the
   *tex_directory* .
#. If a Latex error is encountered, the pdf build will stop with a message
   at the ``?`` prompt. If you enter ``q`` at this prompt, it will complete
   its processing in batch mode. You will be able to find the error messages
   in the file *project_name*\ ``.log`` in the *tex_directory* .
#. Translating Latex errors to the corresponding xrst input file:

   #. Latex error messages are reported using line numbers in
      the file *project*\ ``.tex`` .
   #. You may be able to find the corresponding xrst input file
      using by using ``grep`` to find text that is near the error.
   #. The page numbers in the :ref:`xrst_contents-title` are
      present in the latex input (often near ``section*{`` above the error)
      and may help translate these line numbers to page names.
   #. Given a page name, the corresponding xrst input file can
      be found at the top of the html version of the page.
{xrst_comment --------------------------------------------------------------- }

number_jobs
***********
This is a positive integer specifying the number of parallel jobs
that xrst is allowed to use.
The default value for *number_jobs* is ``1`` .

{xrst_comment --------------------------------------------------------------- }

group_list
**********
It is possible to select one or more groups of pages
to include in the output using this argument.

#. The *group_list* is a list of one or more
   :ref:`group names<begin_cmd@group_name>`.
#. The :ref:`begin_cmd@group_name@Default Group` is represented by
   the group name ``default`` .
#. The order of the group names determines their order in the resulting output.
#. The default value for *group_list* is ``default`` .

For each group name in the *group_list*
there must be an entry in :ref:`config_file@root_file` specifying the
root file for that group name.

The xrst examples are a subset of its user documentation
and its user documentation is a subset of its developer documentation.
For each command, the same source code file provides both the
user and developer documentation. In addition, the developer documentation
has links to the user documentation and the user documentation has links
to the examples.

Example
=======
The examples commands below assume you have cloned the
`xrst git repository <https://github.com/bradbell/xrst>`_
and it is your current working directory.

#. The xrst examples use the default group
   and their documentation can be built using

      ``xrst --group_list default``

#. The xrst user documentation uses the default and user groups
   and its documentation can be built using

      ``xrst --group_list default user``

#. The xrst developer documentation uses the default, user, and dev
   groups and its documentation can be built using

      ``xrst --group_list default user dev``
{xrst_comment --------------------------------------------------------------- }

rename_group
************
If this option is present on the command line,
the :ref:`begin_cmd@group_name` in a subset of the source code, is changed.
This option replaces the :ref:`run_xrst@group_list`
by the list whose only entry is *new_group_name* .
None of the output files are created when rename_group is present;
e.g., the \*.rst and \*.html files.

old_group_name
==============
is the old group name for the pages that will have their group name replaced.
Use ``default``, instead of the empty group name, for the
:ref:`begin_cmd@group_name@Default Group` .

new_group_name
==============
Only the pages below the :ref:`config_file@root_file`
for *new_group_name* are modified.
You can rename a subset of the old group by making the root file
for the new group different than the root file for the old group.
Each page in the old group, and below the root file for the new group,
will have its group name changed from *old_group_name* to *new_group_name*.
Use ``default``, instead of the empty group name, for the
:ref:`begin_cmd@group_name@Default Group` .
{xrst_comment --------------------------------------------------------------- }

{xrst_end run_xrst}
"""
# ---------------------------------------------------------------------------
# imports
# ---------------------------------------------------------------------------
import sys
import re
import os
import toml
import string
import shutil
import filecmp
import argparse
import subprocess
import platform
# ---------------------------------------------------------------------------
# system_exit
# Error messages in this file do not use xrst.system_exit because
# they do not have file names that are relative to the project_directory.
def system_exit(msg) :
   # assert False, msg
   sys.exit(msg)
# ---------------------------------------------------------------------------
# Execute a system command
#
# command:
# the command to execute
#
# warning:
# If a warning is printed, this flag is set True
#
# page_name2line_tuple
# a mapping from page name to to a list of line number pairs.
# The first number is the pair is a line number in the rst file for this page.
# The second number is a corresponding xrst input line number.
#
# page_name2page_file
# map page name to the xrst input file that contains its begin command.
#
def system_command(
      command                    ,
      warning                    ,
      page_name2line_tuple = None ,
      page_name2page_file  = None ,
) :
   assert type(command) == str
   #
   assert type(warning) == list
   assert len(warning) == 1
   assert type(warning[0]) == bool
   #
   if page_name2line_tuple != None :
      assert type(page_name2line_tuple) == dict
      assert type(page_name2page_file) == dict
   #
   # build_link
   build_html = command.startswith('sphinx-build -b html')
   build_tex  = command.startswith('sphinx-build -b tex')
   build_link = command.startswith('sphinx-build -b linkcheck')
   assert build_html or build_tex or build_link
   #
   # subprocess.run, stderr
   print(command)
   command = command.split(' ')
   result = subprocess.run(command, capture_output = True)
   stderr = result.stderr.decode('utf-8')
   #
   # alert_data, pattern_alert
   if build_html or build_tex :
      #
      # alert_data
      alert_data = stderr
      #
      # PAGE_NAME_PATTERN = [A-Za-z0-9._-]+
      # use git grep PAGE_NAME_PATTERN to get all occurances of this pattern
      pattern_alert = re.compile(
         r'.*/rst/([A-Za-z0-9_.-]+).rst:([0-9]+):([^:]*):'
      )
   else :
      assert build_link
      #
      # alert_data
      target_directory = command[-1]
      file_obj         = open( f'{target_directory}/output.txt' , 'r')
      alert_data       = file_obj.read()
      file_obj.close()
      if alert_data == '' :
         alert_data = stderr
      else :
         alert_data = stderr + '\n' + alert_data
      #
      # pattern_alert
      pattern_alert = re.compile( r'^([A-Za-z0-9_.-]+).rst:([0-9]+):' )
   #
   ok  =  result.returncode == 0 and alert_data == ''
   if ok :
      return
   #
   if page_name2line_tuple == None :
      message  = alert_data
      if result.returncode == 0 :
         message  += '\nWarning: see message above.\n'
         sys.stderr.write(message)
         warning[0] = True
         return
      else :
         message  += '\nError: see message above.'
         system_exit(message)
   #
   # message
   message = ''
   #
   # warn_not_xrst_label
   warn_not_xrst_label = True
   #
   # sphinx_error
   sphinx_error = result.returncode != 0
   #
   # alert_list
   alert_list = alert_data.split('\n')
   while '' in alert_list :
      alert_list.remove('')
   if len(alert_list) > 0 :
      if alert_list[-1].strip() == 'warnings.warn(' :
         alert_list.pop(-1)
   #
   # MacOS has a security warning
   if len(alert_list) == 1 :
      issue_320 = 'https://github.com/urllib3/urllib3/issues/3020'
      if alert_list[0].find(issue_320) != -1 :
         uname = platform.uname().system
         if uname == 'Darwin' :
            message  = 'MacOS python has an issue; see '
            message += issue_320 + '\n'
            sys.stderr.write(message)
            return
   #
   for alert in alert_list :
      #
      # m_alert
      m_alert = pattern_alert.search(alert)
      if m_alert == None :
         # this line is more info about previous line
         pass
      else :
         #
         # sphinx_error
         if build_html or build_tex:
            sphinx_error |= m_alert.group(3).strip() == 'ERROR'
         #
         # page_name
         page_name = m_alert.group(1)
         if not page_name in page_name2line_tuple :
            msg  = f'Cannot map line numbers for page {page_name}\n'
            msg += 'to its xrst input file and line number:\n'
            msg += alert
            msg += '\nTry removing the html output directory and re-run.'
            system_exit(msg)
         else :
            #
            # rst_line
            rst_line  = int( m_alert.group(2) )
            #
            # page_file, line_tuple
            page_file  = page_name2page_file[page_name]
            line_tuple = page_name2line_tuple[page_name]
            #
            # n_tuple
            n_tuple = len(line_tuple)
            #
            # index
            index  = 0
            while index < n_tuple and line_tuple[index][0] < rst_line :
                  index += 1
            assert len(line_tuple[index]) == 2 or len(line_tuple[index]) == 4
            #
            # line_before, line_after
            if index == n_tuple :
               line_before = str( line_tuple[n_tuple-1][1] )
               line_after  = '?'
            elif line_tuple[index][0] == rst_line :
               line_before = line_tuple[index][1]
               line_after  = line_tuple[index][1]
            elif 0 == index :
               line_before = '?'
               line_after  = line_tuple[index][1]
            else :
               line_before = line_tuple[index-1][1]
               line_after  = line_tuple[index][1]
            #
            # alert
            msg   = alert[ m_alert.end()  : ]
            if build_html or build_tex :
               msg = m_alert.group(3) + ':' + msg
            if line_before == line_after :
               alert = f'{page_file}:{line_before}:'
            else :
               alert = f'{page_file}:{line_before}-{line_after}:'
            if len( line_tuple[index] ) == 4 :
               template_file = line_tuple[index][2]
               template_line = line_tuple[index][3]
               alert += f'{template_file}:{template_line}:'
            alert += msg
      #
      # alert
      pattern_undefined = re.compile(r"undefined *label: *'([^']*)'")
      m_undefined       = pattern_undefined.search(alert)
      if m_undefined != None :
         label      = m_undefined.group(1)
         xrst_label = 0 < label.find('@')
         if label.endswith('-name') or label.endswith('-title') :
            xrst_label = True
         if not xrst_label and warn_not_xrst_label :
            warn_not_xrst_label = False
            alert += '\n   The label above does not contain an @'
            alert += ' or end with -name or -title.'
            alert += '\n   Hence it is not automatically generated by xrst.'
         if label == 'xrst_table_of_contents-title' :
            alert += '\n   2024-03-04: The label above was changed to'
            alert += " 'xrst_contents-title'"
      #
      # message
      if alert != '' :
         message += '\n' + alert
   #
   if sphinx_error :
      message  += '\nError: see messages above.'
      system_exit(message)
   #
   message  += '\nWarning: see messages above.\n'
   sys.stderr.write(message)
   warning[0] = True
   return
# ---------------------------------------------------------------------------
def fix_latex(latex_dir, project_name) :
   assert type(latex_dir) == str
   assert type(project_name) == str
   #
   # file_name
   file_name = f'{latex_dir}/{project_name}.tex'
   #
   # file_data
   file_obj  = open(file_name, 'r')
   file_data = file_obj.read()
   file_obj.close()
   #
   # file_data
   pattern   = re.compile( r'\n\\section{' )
   file_data = pattern.sub( r'\n\\section*{', file_data)
   #
   # file_data
   pattern   = re.compile( r'\n\\subsection{' )
   file_data = pattern.sub( r'\n\\subsection*{', file_data)
   #
   # file_data
   pattern   = re.compile( r'\n\\subsubsection{' )
   file_data = pattern.sub( r'\n\\subsubsection*{', file_data)
   #
   # file_name
   file_obj  = open(file_name, 'w')
   file_obj.write(file_data)
   file_obj.close()
   #
   return
# ---------------------------------------------------------------------------
# sys.path
# used so that we can test before installing
if( os.getcwd().endswith('/xrst.git') ) :
   if( os.path.isdir('xrst') ) :
      sys.path.insert(0, os.getcwd() )
#
import xrst
#
# version
version = '2025.0.1'
#
def run_xrst() :
   #
   # execution_directory
   execution_directory = os.getcwd()
   #
   # parser
   parser = argparse.ArgumentParser(
      prog='xrst', description='extract Sphinx RST files'
   )
   # --version
   parser.add_argument('--version', action='store_true',
      help='just print version of xrst'
   )
   # --local_toc
   parser.add_argument('--local_toc', action='store_true',
      help='add a local table of contents at the top of each page'
   )
   # --page_source
   parser.add_argument('--page_source', action='store_true',
      help='add link to the xrst source code be included at top of each page'
   )
   # --external_links
   parser.add_argument('--external_links', action='store_true',
      help='report external links that are broken or redirects'
   )
   # --replace_spell_commands
   parser.add_argument('--replace_spell_commands', action='store_true',
      help='replace the xrst spell commands in source code files'
   )
   # --ignore_spell_commands
   parser.add_argument('--ignore_spell_commands', action='store_true',
      help='ignore the xrst spell commands in source code files'
   )
   # --suppress_spell_warnings
   parser.add_argument('--suppress_spell_warnings', action='store_true',
      help='do not generate any of the spell checker wrnings'
   )
   # --continue_with_warning
   parser.add_argument('--continue_with_warnings', action='store_true',
      help='do not exit with an error when warnings are generated'
   )
   # --rst_line_numbers
   parser.add_argument('--rst_line_numbers', action='store_true',
      help='report sphinx errors and warnings using rst file line numbers'
   )
   # --rst_only
   parser.add_argument('--rst_only', action='store_true',
      help='Only extract the sphinx rst files; i.e., do not run sphinx.'
   )
   # --index_page_name
   parser.add_argument(
      '--index_page_name', metavar='index_page_name', default='xrst_root_doc',
      help='The file index.html will be a redirect to this page' + \
         '(default is root of documentation tree)'
   )
   # --config_file
   parser.add_argument(
      '--config_file', metavar='config_file', default='xrst.toml',
      help='location of the xrst configuration file which is in toml format' + \
         '(default is .)'
   )
   # --html_theme
   parser.add_argument(
      '--html_theme', metavar='html_theme', default='furo',
      help='sphinx html_theme that is used to display web pages ' + \
         '(default is furo)',
   )
   # --target
   parser.add_argument(
      '--target', metavar='target', choices=['html', 'tex'], default='html',
      help='type of output files, choices are html or tex (default is html)'
   )
   # --number_jobs
   parser.add_argument(
      '--number_jobs', metavar='number_jobs', default='1',
      help='number of parallel jobs xrst is allowed to use (default is 1)'
   )
   # --group_list
   parser.add_argument(
      '--group_list', nargs='+', default='default',
      metavar= 'group_name' ,
      help='list of group_names to include in this build (default is default)'
   )
   # rename_group
   parser.add_argument(
      '--rename_group', nargs=2, default=None,
      metavar=('old_group_name', 'new_group_name'),
      help='change group name for pages below root page for new_group_name' \
         + '(no default)'
   )
   #
   # arguments
   arguments = parser.parse_args()
   #
   if arguments.version :
      print(version)
      sys.exit(0)
   #
   # local_toc
   local_toc = arguments.local_toc
   #
   # page_source
   page_source = arguments.page_source
   #
   # external_links
   external_links = arguments.external_links
   #
   # index_page_name
   index_page_name = arguments.index_page_name
   #
   # config_file
   # can not use system_exit until os.getcwd() returns project_directory
   config_file = arguments.config_file
   if not os.path.isfile(config_file) :
      msg  = 'xsrst: Error\n'
      msg += f'config_file = {config_file}\n'
      if config_file[0] == '/' :
         msg += 'is not a file\n'
      else :
         msg += f'is not a file relative to the execution directory\n'
         msg += execution_directory
      system_exit(msg)
   #
   # conf_dict
   conf_dict  = xrst.get_conf_dict(config_file)
   #
   # config_dir
   index = config_file.rfind('/')
   if 0 <= index :
      config_dir = config_file[: index]
      os.chdir(config_dir)
   #
   # project_directory
   project_directory = conf_dict['directory']['project_directory']
   #
   # make project directory the current working directory
   os.chdir(project_directory)
   #
   # replace_spell_commands
   replace_spell_commands = arguments.replace_spell_commands
   #
   # ignore_spell_commands
   ignore_spell_commands = arguments.ignore_spell_commands
   #
   # suppress_spell_warnings
   suppress_spell_warnings = arguments.suppress_spell_warnings
   #
   # continue_with_warnings
   continue_with_warnings = arguments.continue_with_warnings
   #
   # rst_line_numbers
   rst_line_numbers = arguments.rst_line_numbers
   #
   # rst_only
   rst_only = arguments.rst_only
   #
   # html_theme
   html_theme = arguments.html_theme
   #
   # number_jobs
   number_jobs = int( arguments.number_jobs )
   if number_jobs <= 0 :
      msg = 'xrst number_jobs is less than or equal zero.'
      system_exit(msg)
   #
   # group_list
   group_list = arguments.group_list
   if type(group_list) == str :
      group_list = [ group_list ]
   #
   # rename_group, group_list
   rename_group = arguments.rename_group
   if rename_group != None :
      group_list = [ rename_group[1] ]
      #
      if rename_group[0] == '' :
         msg = 'xrst rename_group: old_group_name is empty.\n'
         msg += 'Use "default" for the empty group name.'
         system_exit(msg)
      if rename_group[1] == '' :
         msg = 'xrst rename_group: new_group_name is empty.\n'
         msg += 'Use "default" for the empty group name.'
         system_exit(msg)
   #
   # replace_spell_commands or rename_group
   if replace_spell_commands or rename_group != None :
      if replace_spell_commands :
         option = 'replace_spell_commands'
      else :
         option = 'rename_group'
      #
      cwd      = os.getcwd()
      prompt   = f'\nThe {option} option will change some of \n'
      prompt  += 'the files read by xrst. Make sure that you have a backup\n'
      prompt  += f'of source files in {cwd}\n'
      prompt  += 'before contining this operation: continue [yes/no] ? '
      response = None
      while response not in [ 'yes', 'no' ]:
         response = input(prompt)
      if response != 'yes' :
         system_exit( f'xrst: aborting {option}' )
   #
   # target
   target = arguments.target
   #
   # target_directory
   if target == 'html' :
      target_directory = conf_dict['directory']['html_directory']
   else :
      assert target == 'tex'
      target_directory = conf_dict['directory']['tex_directory']
   #
   # rst_directory
   rst_directory = conf_dict['directory']['rst_directory']
   if rst_directory[0] == '/' :
      msg  = 'rst_directory = ' + rst_directory + '\n'
      msg += 'must be a path relative to the project_directory'
      xrst.system_exit(msg)
   #
   if not os.path.isdir(rst_directory) :
      os.makedirs(rst_directory)
   #
   # rst2project_directory
   # relative path from the rst_directory to the project directory
   rst2project_directory = os.path.relpath(
      os.getcwd() , rst_directory
   )
   #
   # tmp_dir
   tmp_dir = rst_directory + '/tmp'
   if os.path.isdir(tmp_dir) :
      shutil.rmtree(tmp_dir)
   os.mkdir(tmp_dir)
   #
   # _sources
   if not os.path.isdir(rst_directory + '/_sources') :
      os.mkdir(rst_directory + '/_sources')
   #
   # spell_checker
   spell_list  = list()
   for entry in conf_dict['project_dictionary']['data'] :
      word_list = entry.split('\n')
      for word in word_list :
         word = word.strip(' \t')
         if len(word) > 0 :
            spell_list.append(word)
   package       = conf_dict['spell_package']['data']
   spell_checker = xrst.get_spell_checker(spell_list, package)
   #
   # not_in_index_list
   not_in_index_list = list()
   for entry in conf_dict['not_in_index']['data'] :
      pattern_list = entry.split('\n')
      for pattern in pattern_list :
         pattern = pattern.strip(' \t')
         try :
            not_in_index_list.append( re.compile(pattern) )
         except :
            msg  = f'not_in_index table in config_file = {config_file}\n'
            msg += f'The regular expression "{pattern}" would not compile'
            system_exit(msg)
   # -------------------------------------------------------------------------
   #
   # input_file_list
   input_file_list = None
   #
   # root_file
   root_file = conf_dict['root_file']
   #
   # project_name
   project_name = conf_dict['project_name']['data']
   #
   # all_page_info
   # This list accumulates over all the group names
   all_page_info  = list()
   #
   # root_page_list
   # Each group has a root secion (in root_file) at the top if its tree.
   root_page_list = list()
   #
   # page_name2line_tuple, page_name2page_file
   # Each rst page name has a corresponding input file and mapping from
   # rst file line nubmers to input file line numbers.
   page_name2line_tuple = dict()
   page_name2page_file  = dict()
   #
   # any_warning
   any_warning = [ False ]
   #
   # unknown_word_dict
   # Each key in this dict is a word not in dictionary or special words
   # The corresponding value is the number of pages it is an error for.
   unknown_word_dict = dict()
   #
   # group_name
   for group_name in group_list :
      #
      # old_group_name, new_group_name
      if rename_group == None :
         old_group_name = group_name
         new_group_name = group_name
      else :
         old_group_name = rename_group[0]
         new_group_name = rename_group[1]
         assert new_group_name == group_name
      #
      if new_group_name not in root_file :
         msg  = f'The group name {new_group_name} is '
         if rename_group == None :
            msg += 'in --group_list\n'
         else :
            msg += 'is new_group_name in --rename_group\n'
         msg += 'but it is not a valid key for the root_file table of\n'
         msg += f'the configuration file {config_file}'
         xrst.system_exit(msg)
      #
      if not os.path.isfile( root_file[new_group_name] ) :
         file_name = root_file[new_group_name]
         msg  = f'The root_file for group_name {new_group_name} is not a file\n'
         msg += f'file name = {file_name}'
         xrst.system_exit(msg)
      #
      # finfo_stack, finfo_done
      # This information is by file, not page
      finfo_stack      = list()
      finfo_done       = list()
      finfo = {
         'file_in'        : root_file[new_group_name],
         'parent_file'    : None,
         'parent_page'    : None,
      }
      finfo_stack.append(finfo)
      #
      while 0 < len(finfo_stack) :
         #
         # finfo
         # pop first element of stack so that order in tex file and
         # table of contents is correct
         finfo  = finfo_stack.pop(0)
         #
         for finfo_tmp in finfo_done :
            if finfo_tmp['file_in'] == finfo['file_in'] :
               msg  = 'The file ' + finfo['file_in']
               msg += '\nis included twice with '
               msg += f'group_name = "{old_group_name}"\n'
               msg += 'Once in ' + finfo_tmp['parent_file'] + '\n'
               msg += 'and again in ' + finfo['parent_file'] + '\n'
               xrst.system_exit(msg)
         finfo_done.append(finfo)
         #
         file_in              = finfo['file_in']
         parent_file          = finfo['parent_file']
         parent_file_page  = finfo['parent_page']
         assert os.path.isfile(file_in)
         #
         # get xrst docuemntation in this file
         file_page_info = xrst.get_file_info(
            all_page_info,
            old_group_name,
            parent_file,
            file_in,
         )
         #
         # root_page_list
         if finfo['parent_file'] == None :
            assert file_in == root_file[new_group_name]
            if file_page_info[0]['is_parent'] :
               n_page = 1
            else :
               n_page = len( file_page_info )
            for i_page in range( n_page ) :
               page_name = file_page_info[i_page]['page_name']
               root_page_list.append(page_name)
         #
         # parent_page_file_in
         # index in all_page_info of parent page for this file
         parent_page_file_in = None
         if file_page_info[0]['is_parent'] :
            parent_page_file_in = len(all_page_info)
         #
         # add this files pages to all_page_info
         for i_page in range( len(file_page_info) ) :
            # ------------------------------------------------------------
            # page_name, page_data, is_parent, begin_line
            page_name  = file_page_info[i_page]['page_name']
            page_data  = file_page_info[i_page]['page_data']
            is_parent  = file_page_info[i_page]['is_parent']
            begin_line = file_page_info[i_page]['begin_line']
            #
            # parent_page
            if is_parent or parent_page_file_in is None :
               parent_page = parent_file_page
            else :
               parent_page = parent_page_file_in
            #
            # all_page_info
            all_page_info.append( {
               'page_name'      : page_name,
               'file_in'        : file_in,
               'parent_page'    : parent_page,
               'begin_line'     : begin_line,
               'in_parent_file' : file_page_info[i_page]['is_child'],
               'end_line'       : file_page_info[i_page]['end_line'],
               'template_list'  : file_page_info[i_page]['template_list'],
            } )
            # ------------------------------------------------------------
            # template command
            # must come before almost all other commands; see its documentation
            page_data = xrst.template_command(page_data, file_in, page_name)
            # -------------------------------------------------------------
            # ref commands
            # Do this before spell checking so spell checking does not have
            # to deal with newlines in ref roles.
            page_data = xrst.ref_command(page_data)
            # -------------------------------------------------------------
            # comment_command
            page_data = xrst.comment_command(page_data)
            # -------------------------------------------------------------
            # spell_command
            # do after suspend and before other commands to help ignore
            # sectons of text that do not need spell checking
            #
            # page_data, any_warning, unknown_word_dict
            page_data, spell_warning, unknown_word_set = xrst.spell_command(
               tmp_dir         = tmp_dir ,
               data_in         = page_data,
               page_file       = file_in,
               page_name       = page_name,
               begin_line      = begin_line,
               ignore_commands = ignore_spell_commands,
               print_warning   = not suppress_spell_warnings,
               spell_checker   = spell_checker,
            )
            if spell_warning :
               assert not suppress_spell_warnings
               any_warning[0] = True
            for word in unknown_word_set :
               if word in unknown_word_dict :
                  unknown_word_dict[word] += 1
               else :
                  unknown_word_dict[word] = 1
            # -------------------------------------------------------------
            # dir commands
            page_data = xrst.dir_command(page_data, rst2project_directory)
            # -------------------------------------------------------------
            # toc commands
            # page_data, finfo_stack, child_page_list, order
            page_data, child_file, child_page_list, order = \
               xrst.toc_commands(
                  is_parent,
                  page_data,
                  file_in,
                  page_name,
                  old_group_name,
            )
            #
            # all_page_info
            all_page_info[-1]['child_order'] = order
            #
            # page_index, finfo_stack
            page_index = len(all_page_info) - 1
            for file_tmp in child_file :
               finfo_stack.append( {
                  'file_in'        : file_tmp,
                  'parent_file'    : file_in,
                  'parent_page'    : page_index,
               } )
            # ------------------------------------------------------------
            # code commands
            page_data = xrst.code_command(
               page_data,
               file_in,
               page_name,
               rst2project_directory,
            )
            # ------------------------------------------------------------
            # literal command
            page_data = xrst.literal_command(
               page_data,
               file_in,
               page_name,
               rst2project_directory,
            )
            # ------------------------------------------------------------
            # process headings
            # pseudo_heading, page_title, keywords
            #
            check_headings = True
            xrst.process_headings(
               check_headings,
               conf_dict,
               local_toc,
               page_data,
               file_in,
               page_name,
               not_in_index_list,
            )
            check_headings = False
            page_data, page_title, pseudo_heading, keywords = \
            xrst.process_headings(
               check_headings,
               conf_dict,
               local_toc,
               page_data,
               file_in,
               page_name,
               not_in_index_list,
            )
            #
            # all_page_info
            # page title is used by table_of_contents
            all_page_info[page_index]['page_title']  = page_title
            all_page_info[page_index]['keywords']    = keywords
            # -------------------------------------------------------------
            # list_children
            # page_name for each of the children of the current page
            if not is_parent :
               list_children = child_page_list
            else :
               assert order in [ 'before', 'after' ]
               list_children = list()
               for i in range( len(file_page_info) ) :
                  if i != i_page :
                     list_children.append(
                        file_page_info[i]['page_name']
                     )
               if order == 'before' :
                  list_children = child_page_list + list_children
               else :
                  list_children = list_children + child_page_list
            # -------------------------------------------------------------
            # process children
            # want this as late as possible so toctree at end of input
            page_data = xrst.process_children(
               page_name,
               page_data,
               list_children,
            )
            # -------------------------------------------------------------
            #
            # line_tuple and file tmp_dir/page_name.rst
            line_tuple = xrst.temporary_file(
               page_source,
               target,
               pseudo_heading,
               file_in,
               tmp_dir,
               page_name,
               page_data,
            )
            #
            # page_name2line_tuple, page_name2page_file
            page_name2line_tuple[page_name] = line_tuple
            page_name2page_file[page_name]  = file_in
      #
      # check_input_files
      if rename_group == None :
         toc_file_set = set()
         for finfo_tmp in finfo_done :
            toc_file_set.add( finfo_tmp['file_in'] )
         input_file_list, file_list_warning = xrst.check_input_files(
            config_file, conf_dict, group_name, toc_file_set, input_file_list
         )
         if file_list_warning :
            any_warning[0] = True
   #
   # index_page_name
   ok = index_page_name == 'xrst_root_doc' or target == 'tex'
   for page_info in all_page_info :
      ok = ok or page_info['page_name'] == index_page_name
   if not ok :
      msg = f'index_page_name = {index_page_name} is not a valid page name.'
      xrst.system_exit(msg)
   #
   # rename_group
   if rename_group != None :
      xrst.rename_group(tmp_dir, rename_group[0], rename_group[1])
      #
      # tmp_dir
      # reset tmp_dir because rmtree is such a dangerous command
      tmp_dir = f'{rst_directory}/tmp'
      shutil.rmtree(tmp_dir)
      print('xrst --rename_group: OK')
      return
   #
   # replace_spell_commands
   if replace_spell_commands :
      xrst.replace_spell(tmp_dir)
      #
      # tmp_dir
      # reset tmp_dir because rmtree is such a dangerous command
      tmp_dir = f'{rst_directory}/tmp'
      shutil.rmtree(tmp_dir)
      print('xrst --replace_spell_commands: OK')
      return
   #
   # unknown_word_dict
   if not suppress_spell_warnings and len(unknown_word_dict) > 0 :
      assert any_warning[0] == True
      msg  = '\nBelow is a list of (word, count) pairs.\n'
      msg += 'Fix low count words using the spell command.\n'
      msg += 'Consider adding high count words to the project dictionary.\n'
      msg += 'If there are a lot of words and they are all correct,\n'
      msg += 'or a lot of word that are not needed (see above),\n'
      msg += 'consider using --replace_spell_commands.\n'
      sys.stderr.write(msg)
      unknown_word_list = sorted(
         unknown_word_dict.items(), key = lambda pair: pair[1]
      )
      line = ''
      for pair in unknown_word_list :
         pair_str = str(pair)
         if len(line) == 0 :
            line = pair_str
         elif len(line + pair_str) > 80 :
            sys.stderr.write(line + '\n')
            line = pair_str
         else :
            line += ' ' + pair_str
      if len(line) > 0 :
         sys.stderr.write(line + '\n')
      sys.stderr.write('\n')
   #
   # ignore_spell_commands
   if ignore_spell_commands :
      return
   #
   # auto_file
   xrst.auto_file(
      conf_dict, html_theme, target, all_page_info, root_page_list
   )
   #
   # not_rst_list
   not_rst_list = [ 'conf.py' , 'xrst_search.js' ]
   # -------------------------------------------------------------------------
   #
   # rst_directory/*.rst
   tmp_list = os.listdir(tmp_dir)
   rst_list = os.listdir(rst_directory)
   for name in tmp_list :
      src = f'{tmp_dir}/{name}'
      des = f'{rst_directory}/{name}'
      if name.endswith('.rst') or name in not_rst_list :
         if name not in rst_list :
               shutil.copyfile(src, des)
         else :
            if not filecmp.cmp(src, des, shallow=False) :
               os.replace(src, des)
   for name in rst_list :
      if name.endswith('.rst') :
         if name not in tmp_list :
            os.remove( f'{rst_directory}/{name}' )
   #
   # tmp_dir
   # reset tmp_dir because rmtree is such a dangerous command
   tmp_dir = f'{rst_directory}/tmp'
   shutil.rmtree(tmp_dir)
   # -------------------------------------------------------------------------
   if rst_only :
      print('xrst rst_only: OK')
      indent = '\n' + 2 * ' '
      txt = ''
      if external_links :
         txt  = f'The following command will check the external links'
         txt += indent
         txt += f'sphinx-build -b linkcheck -j {number_jobs} '
         txt += f'{rst_directory} {target_directory}'
         txt += '\n\n'
      if target == 'html' :
         txt += f'The following commands will build the html:'
         txt += indent
         txt += f'sphinx-build -b html -j {number_jobs} '
         txt += f'{rst_directory} {target_directory}'
         txt += indent
         txt += f'rm -r {target_directory}/_sources'
         txt += indent
         txt += f'mv {rst_directory}/_sources {target_directory}/_sources'
         if html_theme == 'sphinx_rtd_theme' :
            txt += '\nThis will not modify the sphinx_rtd_theme maximum width'
      else :
         assert target == 'tex'
         latex_dir = f'{target_directory}'
         txt += f'The following command will build the tex:'
         txt += indent
         txt += f'sphinx-build -b latex -j {number_jobs} '
         txt += f'{rst_directory} {latex_dir}'
         txt += '\n'
         txt += f'The following command will build the pdf from the tex:'
         txt += indent
         txt += f'make -C {project_directory}/{latex_dir} {project_name}.pdf'
      print(txt)
      return
   #
   # -------------------------------------------------------------------------
   if external_links :
      command  = f'sphinx-build -b linkcheck -j {number_jobs} '
      command += f'{rst_directory} {target_directory}'
      if rst_line_numbers :
         system_command(command, any_warning)
      else :
         system_command(
            command, any_warning, page_name2line_tuple, page_name2page_file
      )
   if target == 'html' :
      command  = f'sphinx-build -b html -j {number_jobs} '
      command += f'{rst_directory} {target_directory}'
      if rst_line_numbers :
         system_command(command, any_warning)
      else :
         system_command(
            command, any_warning, page_name2line_tuple, page_name2page_file
         )
         #
         # target_directory/_sources
         # replace sphinx _sources directory with proper xrst sources
         if not rst_line_numbers :
            src_dir = f'{rst_directory}/_sources'
            des_dir = f'{target_directory}/_sources'
            print( f'rm -r {des_dir}' )
            shutil.rmtree(des_dir)
            print( f'cp -r {src_dir} {des_dir}' )
            shutil.copytree(src_dir, des_dir)
         #
         # target_directory/index.html
         src_file = f'{target_directory}/{index_page_name}.html'
         des_file = f'{target_directory}/index.html'
         shutil.copyfile(src_file, des_file)
         #
         # target_directory/xrst_search.js
         src_file = f'{rst_directory}/xrst_search.js'
         des_file = f'{target_directory}/xrst_search.js'
         shutil.copyfile(src_file, des_file)
         #
   else :
      assert target == 'tex'
      #
      latex_dir = f'{target_directory}'
      command  = f'sphinx-build -b latex -j {number_jobs} '
      command += f'{rst_directory} {latex_dir}'
      if rst_line_numbers :
         system_command(command, any_warning)
      else :
         system_command(
            command, any_warning, page_name2line_tuple, page_name2page_file
         )
      #
      # latex_dir/project_name.tex
      fix_latex(latex_dir, project_name)
      #
      print('The following command will build the pdf from the latex:')
      print( f'   make -C {project_directory}/{latex_dir} {project_name}.pdf' )
   # -------------------------------------------------------------------------
   # target_directory/_static/css/theme.css
   # see https://stackoverflow.com/questions/23211695/
   #  modifying-content-width-of-the-sphinx-theme-read-the-docs
   if html_theme == 'sphinx_rtd_theme' and target == 'html' :
      pattern = dict()
      pattern['content'] = re.compile(
         r'([.]wy-nav-content[{][^}]*;max-width):[^;]*;'
      )
      pattern['sidebar'] = re.compile(
         r'([.]wy-nav-side[{][^}]*;width):[^;]*;'
      )
      pattern['search'] = re.compile(
         r'([.]wy-side-nav-search[{][^}]*;width):[^;]*;'
      )
      new_value = { 'content':'100%', 'sidebar':'250px', 'search':'250px' }
      file_name = f'{target_directory}/_static/css/theme.css'
      if os.path.exists(file_name + '.bak' ) :
        shutil.copyfile(file_name + '.bak', file_name)
      else :
        shutil.copyfile(file_name, file_name + '.bak')
      try :
         file_obj  = open(file_name, 'r')
         ok        = True
      except :
         ok        = False
      if ok :
         data_in   = file_obj.read()
         file_obj.close()
         match     = dict()
         for key in pattern :
            match[key] = pattern[key].search(data_in)
            ok         = match[key] != None
      if ok :
         for key in pattern :
            match[key] = pattern[key].search(data_in, match[key].end())
            ok        = match[key] == None
      if ok :
         data_out = data_in
         for key in pattern :
            data_tmp  = data_out
            value     = new_value[key]
            data_out  = pattern[key].sub( f'\\1:{value};', data_tmp)
            ok        = data_out != data_tmp
      if not ok :
         msg       = 'warning: cannot modify widths in sphinx_rtd_theme\n'
         sys.stderr.write(msg)
         any_warning[0] = True
      else :
         file_obj  = open(file_name, 'w')
         file_obj.write(data_out)
         file_obj.close()
   # -------------------------------------------------------------------------
   if any_warning[0] and not continue_with_warnings :
      system_exit('xrst: See warnings above')
   else :
      print('xrst: OK')



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin auto_file user}
{xrst_spell
   conf
   js
}

Automatically Generated Files
#############################

#. These files are located in the :ref:`config_file@directory@rst_directory` .
#. A new version of these files is created each time ``xrst`` is run.
#. The files in the ``rst`` subdirectory that do not change are not replaced.
   (This reduces the amount of processing that sphinx needs to do.)
#. All these files have the ``.rst`` extension with the exception of
   conf.py and xrst_search.js .


conf.py
*******
This is the sphinx configuration_ file.
It is the only file in the *rst_directory* directory that is
not an rst file.

.. _configuration:  http://www.sphinx-doc.org/en/master/config


xrst_root_doc.rst
*****************
This file is the top of the documentation tree
built by sphinx. It is one level above the first page in
:ref:`config_file@root_file`.

xrst_contents.rst
*****************
This file contains the table of contents for the last run of ``xrst``.
You can link to the corresponding
page name :ref:`xrst_contents-name`
or page title :ref:`xrst_contents-title`
with the following commands::

   :ref:`xrst_contents-name`
   :ref:`xrst_contents-title`

xrst_index.rst
**************
If :ref:`run_xrst@target` is html,
this file contains a link to the table of contents generated by sphinx.
You can link to the corresponding page name or page title
with the following commands::

   :ref:`xrst_index-name`
   :ref:`xrst_index-title`

The xrst_index.rst file is not created when target is tex.

xrst_search.rst
***************
If :ref:`run_xrst@target` is html,
this file contains the xrst search utility for the last run of ``xrst``.
A link to this search utility appears at top left side of each web page.
You can place a link to this page name :ref:`xrst_search-name`
or page title :ref:`xrst_search-title` using the following commands::

   :ref:`xrst_search-title`
   :ref:`xrst_search-name`

This page tile will start with the :ref:`config_file@project_name`
(which for this documentation is ``xrst`` ).
The xrst_search.rst file is not created when target is tex.

xrst_search.js
**************
This file is used by xrst_search.rst.


{xrst_end auto_file}
"""
# ----------------------------------------------------------------------------
import re
import os
import toml
import xrst
#
# conf_py_constant
conf_py_constant = '''#
# General configuration
extensions = [
   'sphinx.ext.mathjax',
   'sphinx_copybutton',
]
root_doc = 'xrst_root_doc'
#
# This does not seem to work so run_xrst.py copies xrst_search.js after
# target directory directory is created.
# html_js_files = [
# 'xrst_search.js',
'''
#
# xrst_index_rst
xrst_index_rst=r'''.. _xrst_index-name:

!!!!!
index
!!!!!

.. _xrst_index-title:

Link To Index Page
##################
:ref:`Link to index <genindex>`

'''
#
#
# xrst_search_rst_template
# Do not start with a newine because it will get removed from
# test_rst/xrst_search.rst during the check for invisible white space.
xrst_search_rst_template = r'''.. _xrst_search-name:

!!!!!!
search
!!!!!!

.. _xrst_search-title:

{XRST_SEARCH_TITLE}

.. raw:: html

   <noscript><h1>
   This search utility requires Javascript to be enabled
   and Javascript is disabled in your browser.
   </h1></noscript>
   <form name='search'>
   <p><table>
      Keywords include the page name, and words in the title or headings.
      Enter Keywords separated by spaces to reduce the match count:
      <tr><td>
         Keywords
      </td><td>
         Match
      </td><tr><td>
         <input
            type='text'
            name='keywords'
            onkeydown='update_match()'
            size='50'
         ></input>
      </td><td>
         <input
            type='text'
            name='match'
            size=5
         ></input>
      </td></tr>
   </table></p>
   <p><table>
      Selecting a page name or ttle below will go to that page:
      <tr><td>
         Page Name
      </td><td>
         Page Title
      </td></tr>
      <tr><td>
         <textarea
            name='name_match'
            rows='20'
            cols='15'
            onclick='select_match(this)'
            ondblclick='goto_match(this)'
            onkeydown='page_or_title_entry(this)'
         ></textarea>
      </td><td>
         <textarea
            name='title_match'
            rows='20'
            cols='50'
            onclick='select_match(this)'
            ondblclick='goto_match(this)'
            onkeydown='page_or_title_entry(this)'
         ></textarea>
      </td></tr>
   </table></p>
   </form>
   <script type='text/javascript' src='xrst_search.js'>
   </script>
'''
#
# xrst_search_js_constant
xrst_search_js_constant = r'''
var max_match_display_global  = 100;
var keywords_length_global    = -1;
var match2page_name_global    = [];
Initialize();
//
// initialize
function Initialize()
{  update_match();
   document.search.keywords.focus()
}
// update_match
function update_match()
{
   // keywords_value, keywords_length_global
   var keywords_value = document.search.keywords.value;
   if( keywords_value.length == keywords_length_global )
      return;
   keywords_lenght_global = keywords_value.length;
   //
   // keyword_list, n_keyword
   var keyword_list = keywords_value.match(/\S+/g);
   var n_keyword    = 0;
   if( keyword_list != null )
      n_keyword = keyword_list.length;
   //
   // pattern
   var pattern = new Array(n_keyword);
   for(var j = 0; j < n_keyword; j++)
      pattern[j] = new RegExp(keyword_list[j], 'i')
   //
   // name_match, title_match, match2page_name_global, count_match
   var count_match = 0;
   var name_match  = '';
   var title_match = '';
   var n_page      = all_page_info_global.length;
   for(i = 0; i < n_page; i++)
   {
      var match = true;
      for(j = 0; j < n_keyword; j++)
      {  flag  = pattern[j].test( all_page_info_global[i].keywords );
         match = match && flag;
      }
      if( match )
      {
         if( count_match < max_match_display_global )
         {  var name    = all_page_info_global[i].page_name;
            var title   = all_page_info_global[i].page_title;
            name_match  = name_match   + name  + '\n';
            title_match = title_match + title + '\n';
            match2page_name_global[count_match] = name;
         }
         count_match = count_match + 1;
      }
   }
   //
   // document.search.match
   document.search.match.value    = count_match;
   document.search.match.readOnly = true;
   document.search.match.setAttribute('wrap', 'off');
   //
   // document.search.name_match
   document.search.name_match.value    = name_match;
   document.search.name_match.readOnly = false;
   document.search.name_match.setAttribute('wrap', 'off');
   //
   // document.search.title_match
   document.search.title_match.value    = title_match;
   document.search.title_match.readOnly = false;
   document.search.title_match.setAttribute('wrap', 'off');
}
// select_match
function select_match(textarea)
{  // start_pos
   var start_select = textarea.value.substring(0, textarea.selectionStart);
   var start_pos    = Math.max(0, start_select.lastIndexOf('\n') + 1);
   //
   // end_pos
   var length       = textarea.length;
   var end_select   = textarea.value.substring(textarea.selectionEnd, length);
   var end_pos      = textarea.selectionEnd + end_select.indexOf('\n');
   //
   // textarea
   textarea.selectionStart = start_pos;
   textarea.selectionEnd   = end_pos;
}
// goto_match
function goto_match(textarea)
{  var start_select = textarea.value.substring(0, textarea.selectionStart);
   var match_index  = start_select.split('\n').length - 1;
   var page_name    = match2page_name_global[match_index];

   //
   parent.location  = page_name + '.html'
}
// page_or_title_entry()
function page_or_title_entry(textarea)
{  alert( 'Only use single or double click below Page Name and Page Title');
   parent.location  = 'xrst_search.html'
}
'''
#
# ----------------------------------------------------------------------------
# {xrst_begin auto_file_dev dev}
# {xrst_spell
#     conf
#     genindex
#     js
# }
# {xrst_comment_ch #}
#
# Create the automatically generated files
# ########################################
#
# conf_dict
# *********
# is a python dictionary representation of the configuration file.
#
# rst_dir
# =======
# we use *rst_dir* to denote *conf_dict* ['directory']['rst_directory'] .
#
# tmp_dir
# =======
# we use *tmp_dir* to denote *rst_dir*\ /tmp .
# This is the directory where xrst creates a temporary copy of *rst_dir* .
# These files are also automatically generated.
#
# html_theme
# **********
# The html_theme as on the xrst command line.
#
# target
# ******
# is html or tex
#
# all_page_info
# *************
# is a list with length equal to the number of pages.
# with the following key, value pairs (all the keys are strings):
#
# .. csv-table::
#     :header: key, value
#
#     page_name, (str) containing the name of this page.
#     page_title,  (str) containing the title for this page.
#     parent_page, (int) index in all_page_info for the parent of this page.
#     in_parent_file, (bool) is this page in same input file as its parent.
#     keywords, (str) space separated list of index entries for this page.
#     file_in, (str) name of the input file for this page
#     begin_line, (int) line number where begin command is for this page
#     end_line, (int) line number where end command is for this page
#     template_list, (list of str) name of template files used by this page
#
# root_page_list
# **************
# is a list of the root page names (one for each group) in the order
# they will appear in the table of contents.
#
# tmp_dir/xrst_contents.rst
# *************************
# This file creates is the table of contents for the documentation.
# It has the label xrst_contents which can be used to link
# to this page.
#
# tmp_dir/xrst_index.rst
# **********************
# This file just contains a link to the genindex.rst file.
# It is (is not) included if target is html (tex).
#
# tmp_dir/xrst_search.rst
# ***********************
# This file contains the xrst search utility.
# It is (is not) included if target is html (tex).
#
# tmp_dir/xrst_search.js
# **********************
# This file contains the java script used by xrst_search.rst.
# It is (is not) included if target is html (tex).
#
# tmp_dir/xrst_root_doc.rst
# *************************
# This is the root level in the sphinx documentation tree.
#
# rst_dir/_sources
# ****************
# The sub-directory is used to store the replacement
# for the _sources directory. This contains the xrst sources that were used
# to create the rst files that sphinx used as sources.
# This is (is not) included if target is html (tex).
# If target is html, this sub-directory must exist and should be empty,
# before calling auto_file.
#
# rst_dir/conf.py
# ***************
# This is the configuration file used by sphinx to build the documentation.
#
#
# Prototype
# *********
# {xrst_literal
#  # BEGIN_DEF
#  # END_DEF
#  }
#
# {xrst_end auto_file_dev}
#
# BEGIN_DEF
def auto_file(
   conf_dict, html_theme, target, all_page_info, root_page_list
   ) :
   assert type(conf_dict) == dict
   assert type(html_theme) == str
   assert type(target) == str
   assert type(all_page_info) == list
   assert type(root_page_list) == list
   # END_DEF
   #
   # rst_dir
   rst_dir = conf_dict['directory']['rst_directory']
   #
   # tmp_dir
   tmp_dir = f'{rst_dir}/tmp'
   #
   # project_name
   project_name = conf_dict['project_name']['data']
   #
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_contents.rst
   #
   # file_data
   file_data     = xrst.table_of_contents(
      tmp_dir, target, all_page_info, root_page_list
   )
   #
   # xrst_contents.rst
   file_out    = tmp_dir + '/' + 'xrst_contents.rst'
   file_obj    = open(file_out, 'w')
   file_obj.write(file_data)
   file_obj.close()
   #
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_index.rst
   if target == 'html' :
      #
      # file_data
      file_data  = xrst_index_rst
      #
      # xrst_index.rst
      file_out    = tmp_dir + '/' + 'xrst_index.rst'
      file_obj    = open(file_out, 'w')
      file_obj.write(file_data)
      file_obj.close()
   #
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_search.rst
   # tmp_dir/xrst_search.js
   if target == 'html' :
      #
      # xrst_search.rst
      file_data   = xrst_search_rst_template
      pattern     = '{XRST_SEARCH_TITLE}'
      line        = f'{project_name} Keyword Search'
      replace     = line + '\n' + len(line) * '*'
      file_data   = re.sub(pattern, replace, file_data)
      file_out    = tmp_dir + '/' + 'xrst_search.rst'
      file_obj    = open(file_out, 'w')
      file_obj.write(file_data)
      file_obj.close()
      #
      # xrst_search.js
      file_data = 'var all_page_info_global = [\n{\n'
      for index, page in enumerate( all_page_info ) :
         for key in [ 'page_name', 'page_title', 'keywords' ] :
            value      = page[key]
            value      = value.replace("'", '"')
            if key == 'keywords' :
               file_data += f"'{key}' : ' {value} '"
            else :
               file_data += f"'{key}' : '{value}'"
               file_data += ',\n'
         if index == len( all_page_info ) - 1 :
            file_data += '\n}\n];'
         else :
            file_data += '\n},{\n'
      file_data  += xrst_search_js_constant
      file_out    = tmp_dir + '/' + 'xrst_search.js'
      file_obj    = open(file_out, 'w')
      file_obj.write(file_data)
      file_obj.close()
   # ------------------------------------------------------------------------
   # rst_dir/_sources
   if target == 'html' :
      #
      # file_in, file_daa
      file_in   = None
      file_data = None
      for info in all_page_info :
         if info['file_in'] != file_in :
            #
            # file_in, file_data
            file_in    = info['file_in']
            file_obj   = open(file_in, 'r')
            file_data  = file_obj.read()
            file_obj.close()
            #
            # newline_list
            newline_list = xrst.newline_indices(file_data)
         #
         # begin_index
         begin_line  = info['begin_line']
         assert 1 <= begin_line
         if begin_line == 1 :
            begin_index = 0
         else :
            assert begin_line < len(newline_list)
            begin_index = newline_list[begin_line-2] + 1
         #
         # end_index
         end_line  = info['end_line']
         assert end_line - 1 <= len(newline_list)
         if end_line - 1 == len(newline_list) :
            end_index = len(file_data)
         else :
            end_index = newline_list[end_line-1] + 1
         #
         # page_data
         title     = f'lines {begin_line}-{end_line} of file: {file_in}'
         dash_line = len(title) * '-'
         title     = f'{dash_line}\n{title}\n{dash_line}\n\n'
         page_data = title + file_data[begin_index : end_index]
         #
         # file_out
         page_name = info['page_name']
         file_out  = rst_dir + '/_sources/' + page_name
         if not page_name.endswith('.rst') :
            file_out += '.rst'
         file_out += '.txt'
         file_obj = open(file_out, 'w')
         file_obj.write(page_data)
         for temp_file in info['template_list'] :
            file_temp  = open(temp_file, 'r')
            data_temp = file_temp.read()
            file_temp.close()
            title     = f'xrst template file: {temp_file}'
            dash_line = len(title) * '-'
            title     = f'\n{dash_line}\n{title}\n{dash_line}\n\n'
            data_temp = title + data_temp
            file_obj.write(data_temp)
         file_obj.close()
   # ------------------------------------------------------------------------
   # rst_dir/conf.py
   #
   # html_theme_options
   if html_theme not in conf_dict['html_theme_options'] :
      html_theme_options = None
   else :
      html_theme_options = conf_dict['html_theme_options'][html_theme]
      html_theme_optiosn = str( html_theme_options )
      assert '"' not in html_theme_options
   #
   # conf_py
   conf_py  = '# This file was automatically generated by xrst\n'
   conf_py += '#\n'
   conf_py += '# Project information\n'
   conf_py += f"project = '{project_name}'\n"
   conf_py +=  "author = ''\n"
   conf_py += conf_py_constant
   conf_py += f'html_theme = "{html_theme}"\n'
   if html_theme_options != None :
      conf_py += f'html_theme_options = {html_theme_options}\n'
   #
   # rst_epilog, rst_prolog, latex_macro
   rst_epilog  = conf_dict['include_all']['rst_epilog']
   rst_prolog  = conf_dict['include_all']['rst_prolog']
   latex_macro = conf_dict['include_all']['latex_macro']
   #
   # rst_prolog
   if target == 'html' and len(latex_macro) != 0 :
      if rst_prolog != '' :
         rst_prolog += '\n\n'
      rst_prolog += '.. rst-class:: hidden\n\n'
      for macro in latex_macro :
         rst_prolog += 2 * ' ' + f':math:`{macro}`\n'
   #
   # conf_py
   if rst_epilog != '' :
      conf_py += '#\n'
      conf_py += f"rst_epilog = r'''\n"
      conf_py += rst_epilog
      conf_py += "'''\n"
   if rst_prolog != '' :
      conf_py += '#\n'
      conf_py += f"rst_prolog = r'''\n"
      conf_py += rst_prolog
      conf_py += "'''\n"
   if len( latex_macro ) != 0 :
      conf_py += '#\n'
      conf_py += '# Latex used when sphinx builds tex\n'
      conf_py += 'latex_elements = {\n'
      conf_py += 2 * ' ' + "'preamble' : r'''\n"
      for macro in latex_macro :
         conf_py += 2 * ' ' + macro + '\n'
      conf_py +=  2 * ' ' +"'''\n"
      conf_py += '}\n'
   #
   # rst_dir/conf.py
   file_out    = rst_dir + '/conf.py'
   file_obj    = open(file_out, 'w')
   file_obj.write(conf_py)
   file_obj.close()
   # ------------------------------------------------------------------------
   # tmp_dir/xrst_root_doc.rst
   #
   # file_data
   file_data  = '.. comment This file was automatically generated by xrst\n\n'
   file_data += project_name + '\n'
   file_data += len(project_name) * '#'
   file_data += '\n\n'
   file_data += '.. toctree::\n'
   file_data += 2 * ' ' + ':maxdepth: 1\n'
   file_data += '\n'
   if target == 'html' :
      file_data += 2 * ' ' + 'xrst_search\n'
      file_data += 2 * ' ' + 'xrst_index\n'
   file_data += 2 * ' ' + 'xrst_contents\n'
   for page_name in root_page_list :
      file_data += 2 * ' ' + '' + page_name + '\n'
   #
   # xrst_root_doc.rst
   file_name    = tmp_dir + '/xrst_root_doc.rst'
   file_obj     = open(file_name, 'w')
   file_obj.write(file_data)
   file_obj.close()



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin indent_cmd user}

Indent Command
##############


Syntax
******
*indentation* ``\{xrst_indent}``

Discussion
**********
#. The xrst documentation for a page can be indented so that it is grouped
   with the source code it is next to.
#. We call lines that have just spaces or tabs empty lines
#. Empty lines below an indent command, and before the next indent command,
   must begin with *indent* . The indent characters
   are not included in the rst output; i.e., the indentation is removed.
#. If there is an indent command in a page,
   lines before the first indent command do not have any indentation.
#. If there is no indent command in a page,
   the indentation for the page is computed automatically.

indentation
***********
This is the sequence of spaces or a sequence of tabs that
come before ``\{xrst_indent}`` .
It cannot mix both spaces and tabs.

Automatic Indentation
*********************
If there is no indent command in a page,
the indentation is calculated automatically as follows:
The number of spaces (or tabs) before
all of the xrst documentation for a page
(not counting lines with just spaces or tabs)
is used for the indentation for that page.

Comment Character
*****************
The indentation is calculated before processing the
:ref:`comment_ch_cmd-name` command.

Example
*******
:ref:`indent_example-name`,
:ref:`comment_ch_example@Indent`, and
:ref:`example_ad_double@xrst_indent` .

{xrst_end indent_cmd}
"""
import re
import xrst
# {xrst_begin indent_cmd_dev dev}
# {xrst_comment_ch #}
#
# Process indent commands in a page
# #################################
#
# Prototype
# *********
# {xrst_literal ,
#  # BEGIN_DEF, # END_DEF
#  # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page.
#
# file_name
# *********
# is the input that this page appears in (used for error reporting).
#
# page_name
# *********
# is the name of this page (used for error reporting).
#
# data_out
# ********
# is a copy of data_in with the indentation for this section removed.
#
# {xrst_end indent_cmd_dev}
#

# pattern_indent
# 0. newline, characters before comamnd, command
# 1. indent; i.e. , characters on same line and before the command.
pattern_indent = re.compile(
   r'\n{xrst_indent}|\n([^\n{]*[^\n\\]){xrst_indent}'
)
#
# pattern_newline
pattern_newline = re.compile( r'\n([ \t]*)[^ \t\n]' )
#
# pattern_space_or_tab
pattern_space_or_tab = re.compile( r'[ \n]*' )
#
# BEGIN_DEF
def indent_command(data_in, file_name, page_name) :
   assert type(data_in) == str
   assert type(file_name) == str
   assert type(page_name) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_obj
   m_obj = pattern_indent.search(data_out)
   if m_obj == None :
      #
      # indentation
      indentation = xrst.auto_indent(data_in, file_name, page_name)
      #
      # data_out
      pattern  = re.compile( r'\n' + indentation )
      data_out = pattern.sub('\n', data_out )
   else:
      #
      # m_obj
      while m_obj != None :
         #
         # indentation
         indentation = m_obj.group(1)
         if indentation == None :
            indentation = ''
         if pattern_space_or_tab.fullmatch(indentation) == None :
            msg  = f'indentation before {xrst_indent} = "'
            msg += indent + '"\n'
            msg += 'has characters that are not spaces or tabs'
            xrst.system_exit(msg,
               file_name = file_name,
               page_name = page_name,
               m_obj     = m_obj,
               data      = data_out
            )
         if '\n' in indentation and ' ' in indentation :
            msg = 'indentation before {xrst_indent} has spaces and tabs'
            xrst.system_exit(msg,
               file_name = file_name,
               page_name = page_name,
               m_obj     = m_obj,
               data      = data_out
            )
         #
         # this_start, this_end
         this_start  = m_obj.start()
         this_end    = m_obj.end()
         #
         # next_start, next_end
         m_next = pattern_indent.search(data_out, m_obj.end())
         if m_next == None :
            next_start = len(data_out)
         else :
            next_start = m_next.start()
         #
         # data_before, data_middle, data_after
         # include first match character (newline) in data_before
         data_before = data_out[: this_start + 1]
         data_middle = data_out[this_end : next_start]
         data_after  = data_out[next_start :]
         #
         # m_check
         for m_check in pattern_newline.finditer(data_middle) :
            if not m_check.group(1).startswith(indentation) :
               found = m_check.group(1)
               msg  = 'line does not begin with indentation '
               msg +='in previous indent command\n'
               msg += f'expected = "{indentation}"\n'
               msg += f'found    = "{found}"'
               xrst.system_exit(msg,
                  file_name = file_name,
                  page_name = page_name,
                  m_obj     = m_check,
                  data      = data_middle
               )
         #
         # data_middle
         pattern     = re.compile( r'\n' + indentation )
         data_middle = pattern.sub('\n', data_middle)
         #
         # data_left, data_out
         data_left = data_before + data_middle
         data_out  = data_left + data_after
         #
         # m_obj
         m_obj = pattern_indent.search(data_out, len(data_left) )
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
#
# {xrst_begin module dev}
# {xrst_comment_ch #}
#
# The xrst Module
# ###############
#
# {xrst_comment BEGIN_SORT_THIS_LINE_PLUS_2}
# {xrst_toc_table
#  xrst/add_before_title.py
#  xrst/add_line_numbers.py
#  xrst/auto_file.py
#  xrst/auto_indent.py
#  xrst/check_input_files.py
#  xrst/check_page_name.py
#  xrst/code_command.py
#  xrst/comment_command.py
#  xrst/dir_command.py
#  xrst/file_line.py
#  xrst/get_conf_dict.py
#  xrst/get_file_info.py
#  xrst/get_spell_checker.py
#  xrst/indent_command.py
#  xrst/literal_command.py
#  xrst/newline_indices.py
#  xrst/next_heading.py
#  xrst/pattern.py
#  xrst/process_children.py
#  xrst/process_headings.py
#  xrst/ref_command.py
#  xrst/remove_line_numbers.py
#  xrst/rename_group.py
#  xrst/replace_spell.py
#  xrst/spell_command.py
#  xrst/sphinx_label.py
#  xrst/start_end_file.py
#  xrst/suspend_command.py
#  xrst/system_exit.py
#  xrst/table_of_contents.py
#  xrst/template_command.py
#  xrst/temporary_file.py
#  xrst/toc_commands.py
# }
# {xrst_comment END_SORT_THIS_LINE_MINUS_2}
#
# {xrst_end module}
#
# Must import pattern first because it is used by some of the other imports
from .pattern                import pattern
#
# BEGIN_SORT_THIS_LINE_PLUS_1
from .add_before_title       import add_before_title
from .add_line_numbers       import add_line_numbers
from .auto_file              import auto_file
from .auto_indent            import auto_indent
from .check_input_files      import check_input_files
from .check_page_name        import check_page_name
from .code_command           import code_command
from .comment_ch_command     import comment_ch_command
from .comment_command        import comment_command
from .dir_command            import dir_command
from .file_line              import file_line
from .get_conf_dict          import get_conf_dict
from .get_file_info          import get_file_info
from .get_spell_checker      import get_spell_checker
from .indent_command         import indent_command
from .literal_command        import literal_command
from .newline_indices        import newline_indices
from .next_heading           import next_heading
from .process_children       import process_children
from .process_headings       import process_headings
from .ref_command            import ref_command
from .remove_line_numbers    import remove_line_numbers
from .rename_group           import rename_group
from .replace_spell          import replace_spell
from .run_xrst               import run_xrst
from .spell_command          import spell_command
from .sphinx_label           import sphinx_label
from .start_end_file        import start_end_file
from .suspend_command        import suspend_command
from .system_exit            import system_exit
from .table_of_contents      import table_of_contents
from .template_command       import template_command
from .temporary_file         import temporary_file
from .toc_commands           import toc_commands
# END_SORT_THIS_LINE_MINUS_1



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin code_example}

Code Command Example
####################

Factorial
*********
{xrst_code py}"""
def factorial(n) :
   if n == 1 :
      return 1
   return n * factorial(n-1)
r"""{xrst_code}

Git Hash
********
{xrst_spell_off}
{xrst_code py}"""
git_hash='7c35a3ce607a14953f070f0f83b5d74c2296ef93'
r"""{xrst_code}
{xrst_spell_on}

xrst_code
*********
The file below demonstrates the use of ``xrst_code`` .

xrst_spell_off, xrst_spell_on
*****************************
The file below demonstrates the use of ``xrst_spell_off``
and ``xrst_spell_on`` .

This Example File
*****************
{xrst_literal}

{xrst_end code_example}
"""



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
# indent_pattern
indent_pattern = re.compile( r'^\n[ \t]*' )
#
# line_number_pattern
line_number_pattern = xrst.pattern['line']
#
# {xrst_begin add_line_numbers dev}
# {xrst_comment_ch #}
#
# Add Line Numbers to File Data
# #############################
# Add line numbers to a string in a way that is useful for reporting errors
# (for modified versions of string) using line number in the original string.
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# The original string.  An empty line is a line with just spaces or tabs.
# line_number is the number of newlines before a line plus one; i.e.,
# the first line is number one.
#
# file_in
# *******
# is the file corresponding to data_in (used for error reporting).
#
# data_out
# ********
# The return data_out is a modified version of data_in. The text
#
#  | ``@xrst_line`` *line_number* ``@``
#
# is added at the end of each non-empty line.
# There is one space between ``@xrst_line`` and *line_number*.
# There is no space between *line_number* and ``@`` .
# Spaces and tabs in empty lines are removed (so they are truly empty).
#
# {xrst_end add_line_numbers}
# BEGIN_DEF
def add_line_numbers(data_in, file_in) :
   assert type(data_in) == str
   assert type(file_in) == str
   # END_DEF
   #
   # data_extend
   data_extend = data_in
   if data_extend[-1] != '\n' :
      data_extend += '\n'
   #
   # m_obj
   m_obj = line_number_pattern.search(data_in)
   if m_obj != None :
      line = data_in[: m_obj.start() ].count('\n') + 1
      msg  = 'It is an error for the following text to appear in xrst input:\n'
      msg += '@xrst_line<space><number>@\n'
      msg += 'where <space> is a single space '
      msg += 'and <number> is an integer'
      xrst.system_exit(msg,
         file_name = file_in,
         line      = line,
         m_obj     = m_obj,
         data      = data_in
      )
   #
   # newline_list, line_start
   newline_list = xrst.newline_indices(data_extend)
   if newline_list[0] == 0 :
      line_start = 2
      newline_list .pop(0)
   else :
      line_start = 1
   #
   # data_out, previous
   data_out     = ""
   previous     = 0
   #
   for i in range( len(newline_list) ) :
      #
      # current
      current = newline_list[i]
      assert previous < current
      #
      # line
      line = data_extend[previous : current]
      #
      # empty_line
      if previous == 0 :
         m_obj = indent_pattern.search( '\n' + line )
         empty_line = m_obj.end() == len(line) + 1
      else :
         m_obj = indent_pattern.search( line )
         empty_line = m_obj.end() == len(line)
      #
      # line
      if empty_line :
         line = '\n'
      else :
         line += '@xrst_line ' + str(i + line_start) + '@'
      #
      # data_out, previous
      data_out  += line
      previous = current
   #
   # data_out
   assert previous == len(data_extend) - 1
   data_out += '\n'
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import sys
import os
import re
import xrst
# {xrst_begin file_line dev}
# {xrst_comment_ch #}
#
# Determine Input File and Line
# #############################
#
# {xrst_literal ,
#     # BEGIN_DEF ,    # END_DEF
#     # BEGIN_RETURN , # END_RETURN
# }
#
#
# m_obj
# *****
# is a match object for a location in data.
#
# data
# ****
# is the data for a page including template commands (or expansions
# if the templates commands have been processed).
#
# page_file
# *********
# This is the file where the begin and end commands appear for this page.
#
# page_line
# *********
# is the line number in the *page_file* corresponding to *m_obj*  .
# If *template_file* is None, *m_obj* starts at this line.
# Otherwise the template expansion for *template_file* starts at this line.
#
# template_file
# *************
# is the name of the file for the template expansion where the start of
# *m_obj* is located.  If *m_obj* does not start in a template expansion,
# *template_file* is None.
# In this case, *m_obj* starts in the file where the begin command for this
# page is located.
#
# template_line
# *************
# if *template_file* is None, *template_line* is None.
# Otherwise it is the line number in the *template_file* for this
# corresponding to this expansion.
#
# {xrst_end file_line}
# BEGIN_DEF
def file_line(m_obj, data) :
   assert type(m_obj) == re.Match
   assert type(data) == str
   # END_DEF
   #
   # match_line
   m_line  = xrst.pattern['line'].search( data[m_obj.start() :] )
   assert m_line != None
   match_line = int( m_line.group(1) )
   #
   #
   # template_expansion, begin_index
   begin_index = data[: m_obj.start()].rfind( '@{xrst_template_begin@' )
   end_index   = data[: m_obj.start()].rfind( '@{xrst_template_end}@' )
   template_expansion = end_index < begin_index
   #
   # page_line, template_file, template_line
   if not template_expansion :
      page_line      = match_line
      template_file  = None
      template_line  = None
   else :
      #
      # template_file, expansion_line
      m_temp = xrst.pattern['template_begin'].search( data[begin_index :] )
      assert m_temp != None
      template_file  = m_temp.group(1).strip()
      page_line      = int( m_temp.group(2) )
      template_line  = match_line
   # BEGIN_RETURN
   #
   assert type(page_line) == int
   if template_file == None :
      assert template_line == None
   else :
      assert type(template_file) == str
      assert type(template_line) == int
   return page_line, template_file, template_line
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin ref_cmd user}

Allow Multiple Lines in Sphinx ref Role
#######################################

Syntax
******

| ``:ref:`` ` *target* `
| ``:ref:`` ` *linking_text* ``<``  *target* ``>`` `

Purpose
*******
The xrst program allows one to place a sphinx ``ref`` role target
on multiple lines.
This makes the xrst input more readable
when the headings corresponding to the target are long; see
:ref:`heading_links-name` .

New Lines
*********
Newlines and spaces surrounding newlines are removed  from *target* .

Example
*******
:ref:`ref_example-name`

{xrst_end ref_cmd}
"""
# ----------------------------------------------------------------------------
import re
import os
import xrst
#
# ref_pattern
ref_pattern_one = re.compile( r':ref:`([^<`]+)`' )
ref_pattern_two = re.compile( r':ref:`([^<`]*)<([^>]*)> *`' )
line_pattern    = re.compile(  r'@xrst_line ([0-9]+)@\n *' )
#
#
# {xrst_begin ref_cmd_dev dev}
# {xrst_comment_ch #}
#
# Remove Leading and Trailing White Space From ref Role Targets
# #############################################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page.
#
# data_out
# ********
# The return data_out is a copy of data_in except that white space
# surrounding the target components has been removed .
#
# {xrst_end ref_cmd_dev}
# BEGIN_DEF
def ref_command(data_in) :
   assert type(data_in) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # offset
   offset = 0
   #
   # m_ref
   m_ref  = ref_pattern_one.search(data_out, offset)
   while m_ref != None :
      #
      # data_before, data_after
      data_before  = data_out[: m_ref.start()]
      data_after   = data_out[m_ref.end() :]
      #
      # target
      target = m_ref.group(1)
      #
      # target
      target = line_pattern.sub('', target)
      #
      # data_middle
      data_middle = f':ref:`{target}`'
      #
      # data_out
      data_out = data_before + data_middle + data_after
      #
      # offset
      offset = len( data_before + data_middle )
      #
      # m_ref
      m_ref = ref_pattern_one.search(data_out, offset)
   #
   # offset
   offset = 0
   #
   # m_ref
   m_ref  = ref_pattern_two.search(data_out, offset)
   while m_ref != None :
      #
      # data_before, data_after
      data_before  = data_out[: m_ref.start()]
      data_after   = data_out[m_ref.end() :]
      #
      # linking_text, target
      linking_text = m_ref.group(1)
      target       = m_ref.group(2)
      #
      # target
      target = line_pattern.sub('', target)
      #
      # data_middle
      data_middle = f':ref:`{linking_text}<{target}>`'
      #
      # data_out
      data_out = data_before + data_middle + data_after
      #
      # offset
      offset = len( data_before + data_middle )
      #
      # m_ref
      m_ref = ref_pattern_two.search(data_out, offset)
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import re
import xrst
#
# pattern_toc
pattern_toc = re.compile(
   r'\n{xrst_TOC_(hidden|list|table)}\n'
)
#
# patttern_rst_extension
pattern_rst_extension = re.compile( r'\.rst$' )
# ----------------------------------------------------------------------------
# {xrst_begin process_children dev}
# {xrst_spell
#     toctree
# }
# {xrst_comment_ch #}
#
# Add child information to a page
# ###############################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page after the toc_command function has processed
# the toc commands.
#
# list_children
# *************
# is a list of the page names for the children of this page.
# If this list is empty, data_out is equal to data_in.
#
# data_out
# ********
# The return value data_out has the child information added.
#
#  #. A hidden table of contents (toctree) for the children is added at the
#     end of data_out.
#  #. If the TOC command in data_in is \{xrst_TOC_list} or \{xrst_TOC_table},
#     the corresponding links will replace the command.
#  #. If the child command is \{xrst_TOC_hidden}, the command is removed
#     and no table of links is added.
#  #. If there is no TOC command and list_children is non-empty,
#     the toc_table style is used for the links to the children which are
#     placed at the end of the data_out (before the toctree).
#
# {xrst_end process_children}
# BEGIN_DEF
def process_children(
   page_name,
   data_in,
   list_children,
) :
   assert type(page_name) == str
   assert type(data_in) == str
   assert type(list_children) == list
   if len(list_children) > 0 :
      assert type(list_children[0]) == str
   # END_DEF
   #
   if len(list_children) == 0 :
      m_child = pattern_toc.search(data_in)
      assert m_child is None
      return data_in
   #
   # data_out
   data_out = data_in
   #
   # m_child
   m_child = pattern_toc.search(data_out)
   #
   # page_has_child_command
   page_has_child_command =  m_child != None
   if page_has_child_command :
      #
      # type of toc command
      toc_type = m_child.group(1)
      #
      # There chould be at most one toc command per page created by
      # the xrst.child_command routine
      m_tmp = pattern_toc.search(data_in, m_child.end())
      assert m_tmp == None
      #
      # cmd
      if toc_type ==  'list' :
         cmd = '\n\n'
         for child in list_children :
            cmd += '-  :ref:`' + child + '-title`\n'
         cmd += '\n\n'
      elif toc_type == 'table' :
         cmd  = '\n\n'
         cmd += '.. csv-table::\n'
         cmd += '   :header:  "Name", "Title"\n'
         cmd += '   :widths: auto\n\n'
         for child in list_children :
            cmd += '   "' + child + '"'
            cmd += ', :ref:`' + child + '-title`\n'
      else :
         assert toc_type == 'hidden'
         cmd = '\n'
      #
      # data_out
      data_tmp = data_out[: m_child.start()]
      data_tmp += cmd
      data_tmp += data_out[m_child.end() :]
      data_out  = data_tmp
   #
   # -----------------------------------------------------------------------
   if data_out[-1] != '\n' :
      data_out += '\n'
   #
   # data_out
   # If there is no toc command in this page, automatically generate
   # links to the child pages at the end of the page.
   if not page_has_child_command :
      data_out += '.. csv-table::\n'
      data_out += '   :header: "Child", "Title"\n'
      data_out += '   :widths: 20, 80\n\n'
      for child in list_children :
         data_out += '   "' + child + '"'
         data_out += ', :ref:`' + child + '-title`\n'
      data_out += '\n'
   #
   # data_out
   # put hidden toctree at end of page
   toctree  = '\n.. toctree::\n'
   toctree += 2 * ' ' + ':maxdepth: 1\n'
   toctree += 2 * ' ' + ':hidden:\n\n'
   for child in list_children :
      entry    = pattern_rst_extension.sub('', child)
      toctree += 2 * ' ' + '' + entry + '\n'
   data_out = data_out + toctree
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-22 Bradley M. Bell
# ----------------------------------------------------------------------------
import xrst
xrst.run_xrst()



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
import xrst
# ----------------------------------------------------------------------------
# page_index =
def page_name2index(all_page_info, page_name) :
   for (page_index, info) in enumerate(all_page_info) :
      if info['page_name'] == page_name :
         return page_index
   return None

# ----------------------------------------------------------------------------
# Create the table of contents and replace the '{xrst@before_title}'
# for this page and all its child pages.
#
# tmp_dir
# is the temporary directory where the rst files are written.
#
# target:
# is either 'html' or 'tex'. If target is 'tex',  in the temporary files
# tmp_dir/page_name.rst, the text {xrst@before_title}
# is removed and  page number followed by page name is added to the
# title. The page number includes the counter for each level.
# If target is 'html', {xrst@before_title} is removed without other changes.
#
# count:
# is a list where each element is a non-negative int.
#
# count[-1] - 1 is the number of pages at the level of this page and
# before this page.
#
# count[-2] - 1 is the number of pages at the level of this pages parent and
# before this pages parent.
# ...
#
# If this count is the empty list, this page is the root of the table of
# contents tree.
#
# page_index:
# is the index of this page in all_page_info
#
# all_page_info:
# is a list with length equal to the number of pages.
# The value all_page_info[page_index] is a dictionary for this page
# with the following key, value pairs (all the keys are strings:
# key            value
# page_name   a str continaing the name of this page.
# page_title  a str containing the title for this page.
# parent_page an int index in page_info for the parent of this page.
# in_parent_file True if this page in same input file as its parent.
#
# content:
# The return content is the table of contents entries for this page
# and all the pages below this page.
#
# content =
def page_table_of_contents(
   tmp_dir, target, count, all_page_info, page_index
) :
   assert type(tmp_dir) == str
   assert type(target) == str
   assert type(count) == list
   assert type(all_page_info) == list
   assert type(page_index) == int
   #
   assert target in [ 'html', 'tex' ]
   #
   # page_name, page_title, child_order
   page_name   = all_page_info[page_index]['page_name']
   page_title  = all_page_info[page_index]['page_title']
   child_order = all_page_info[page_index]['child_order']
   #
   # page_number, content
   page_number = ''
   if 0 == len(count) :
      content = ''
   else :
      content = '| '
   if 0 < len(count) :
      assert type( count[-1] ) == int
      for i in range( len(count) - 1 ) :
         content += 2 * ' '
      for (i, c) in enumerate(count) :
         page_number += str(c)
         if i + 1 < len(count) :
            page_number += '.'
   #
   # content
   if len(count) == 0 :
      content  += f':ref:`{page_name}-title`' '\n\n'
   else :
      content  += f':ref:`{page_number}<{page_name}-title>` '
      content  += page_title + '\n'
   #
   # file_name
   # temporary file corresponding to this page name
   if page_name.endswith('.rst') :
      file_name = tmp_dir + '/' + page_name
   else :
      file_name = tmp_dir + '/' + page_name + '.rst'
   #
   # page_data
   file_obj  = open(file_name, 'r')
   page_data = file_obj.read()
   file_obj.close()
   page_data = xrst.add_before_title(
      page_data, target, page_number, page_name
   )
   #
   # file_name
   file_obj  = open(file_name, 'w')
   file_obj.write(page_data)
   file_obj.close()
   #
   # in_parent_file_list, in_toc_cmd_list
   in_parent_file_list = list()
   in_toc_cmd_list   = list()
   for child_index in range( len( all_page_info ) ) :
      if all_page_info[child_index]['parent_page'] == page_index :
         if all_page_info[child_index]['in_parent_file'] :
            in_parent_file_list.append(child_index)
         else :
            in_toc_cmd_list.append(child_index)
   #
   # child_content
   child_content = ''
   child_count   = count + [0]
   if child_order == 'before' :
      child_index_list = in_toc_cmd_list + in_parent_file_list
   else :
      assert child_order == 'after'
      child_index_list = in_parent_file_list + in_toc_cmd_list
   for child_index in child_index_list :
      #
      # child_count
      child_count[-1] += 1
      child_content += page_table_of_contents(
         tmp_dir, target, child_count, all_page_info, child_index
      )
   #
   # content
   content += child_content
   #
   return content
# ----------------------------------------------------------------------------
# {xrst_begin table_of_contents dev}
# {xrst_comment_ch #}
#
# Create the table of contents and Modify Titles
# ##############################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# tmp_dir
# *******
# is the temporary directory where the temporary rst files are written.
#
# target
# ******
# is either 'html' or 'tex'.
#
# tex
# ===
# If target is 'tex',  for each temporary file
# tmp_dir/page_name.rst the text \\n\{xrst\@before_title}
# is removed and the page number followed by the page name is added
# at the front of the title for the page.
# The page number includes the counter for each level.
#
# html
# ====
# If target is 'html',
# \\n\{xrst\@before_title} is removed without other changes.
#
# all_page_info
# *************
# is a list with length equal to the number of pages.
# The value all_page_info[page_index] is a dictionary for this page
# with the following key, value pairs (all the keys are strings):
#
# ..  csv-table::
#     :header: key, value, type
#
#     page_name, contains the name of this page, str
#     page_title,  contains the title for this page, str
#     parent_page, index in all_page_info for the parent of this page, int
#     in_parent_file, is this page in same input file as its parent, bool
#
# root_page_list
# **************
# is a list of strings containing the root page name for each group.
# The order of the root page names determine the order of the groups
# int the table of contents.
#
# content
# *******
# The return content is the table of contents entries for all the pages.
# The following are placed at the beginning of the of content.
#
# 1.  The page name xrst_contents and corresponding label xrst_contents-name
# 2.  The page title Table of Contents and corresponding label
#     xrst_contents-title
#
# {xrst_end table_of_contents}
# BEGIN_DEF
def table_of_contents(
   tmp_dir, target, all_page_info, root_page_list
) :
   assert type(tmp_dir) == str
   assert type(target) == str
   assert target in [ 'html', 'tex']
   assert type(all_page_info) == list
   assert type(all_page_info[0]) == dict
   assert type(root_page_list) == list
   assert type(root_page_list[0]) == str
   # END_DEF
   #
   # content
   content  = '.. _xrst_contents-name:\n\n'
   content += '!!!!!!!!\n'
   content += 'contents\n'
   content += '!!!!!!!!\n\n'
   #
   content += '.. _xrst_contents-title:\n\n'
   content += 'Table of Contents\n'
   content += '*****************\n'
   #
   # content
   if len(root_page_list) == 1 :
      count = []
      page_name  = root_page_list[0]
      page_index = page_name2index(all_page_info, page_name)
      content += page_table_of_contents(
         tmp_dir, target, count, all_page_info, page_index
      )
   else :
      count = [0]
      for page_name in  root_page_list :
         page_index = page_name2index(all_page_info, page_name)
         count[0]     += 1
         content      += page_table_of_contents(
            tmp_dir, target, count, all_page_info, page_index
         )
   #
   # BEGIN_RETURN
   #
   assert type(content) == str
   return content
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin auto_indent dev}

Automatic Indentation
#####################
Compute the indentation at the beginning of every line in *data* .
The characters that can be used to indent are spaces or tabs.
Lines that only have spaces and tabs are not included in this calculation.

Prototype
*********
{xrst_literal ,
   # BEGIN_DEF, # END_DEF
   # BEGIN_RETURN, # END_RETURN
}

data
****
is the data we are computing the indentation for.
The text before the first new_line does not matter.
If you want to include this text, add a newline at the beginning of *data*.

file_name
*********
used for error reporting when *data* mixes spaces and tabs in
the indentation.

page_name
*********
used for error reporting when *data* mixes spaces and tabs in
the indentation.

indentation
***********
The return value *indentation* is the automatically computed indentation.
It will be a sequence of spaces or tabs but it will not mix
spaces and tabs.

{xrst_end auto_indent}
"""
import re
import xrst
#
# pattern_newline
pattern_newline = re.compile( r'\n([ \t]*)[^ \t\n]' )
#
# BEGIN_DEF
def auto_indent(data, file_name, page_name) :
   assert type(data) == str
   assert type(file_name) == str
   assert type(page_name) == str
   # END_DEF
   #
   # len_data
   len_data   = len(data)
   #
   # newline_list
   newline_list = xrst.newline_indices(data)
   #
   # n_indent
   # use match_empty to gaurd against case where there are no matches
   n_indent    = len(data)
   m_itr       = pattern_newline.finditer(data)
   match_empty = True
   for m_obj in m_itr :
      match_empty = False
      n_indent    = min(n_indent, len( m_obj.group(1) ) )
   if match_empty :
      n_indent = 0
   #
   # check if there is no indent to remove
   if n_indent == 0 :
      return ''
   #
   # indent_ch
   line      = 0
   indent_ch = data[ newline_list[line] + 1 ]
   while indent_ch == '\n' :
      line += 1
      indent_ch = data[ newline_list[line] + 1 ]
   #
   # check for mixing spaces and tabs
   check_ch  = indent_ch + '\n'
   for newline in newline_list :
      next_ = newline + 1
      end   = min( len_data, next_ + n_indent )
      while next_ < end :
         if data[next_] not in check_ch :
            msg  = 'mixing both spaces and tabs for '
            msg += 'white space that indents this page.'
            xrst.system_exit(
               msg, file_name=file_name, page_name=page_name
            )
         next_ += 1
   #
   #
   # indentation
   indentation = n_indent * indent_ch
   #
   # BEGIN_RETURN
   assert type(indentation) == str
   return indentation
   # END_RETURN



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-25 Bradley M. Bell
# ----------------------------------------------------------------------------
import os
import setuptools
#
# BEGIN_DEPENDENCIES
dependencies = [
   # Always requires to run xrst
   'sphinx', 'toml', 'sphinx-copybutton',
   # Ony required to test xrst
   'pytest',
   # Only required if used in xrst configure file
   'pyspellchecker', 'pyenchant',
   # Only required if used in xrst configure file
   'furo', 'sphinx-rtd-theme', 'sphinx-book-theme',
]
if os.name == 'nt' :
   # pyenchant is not available on windows conda-forge
   dependencies.remove('pyenchant')
# END_DEPENDENCIES
#
# setuptools.setup
setuptools.setup(
   install_requires = dependencies,
)



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin dir_cmd user}

Converting Sphinx Command File Names
####################################

Syntax
******
``\{xrst_dir`` *file_name* ``}``

Purpose
*******
Sphinx commands that use file names must specify the file
relative to the :ref:`config_file@directory@rst_directory` .
The xrst dir command converts a file name relative to the
:ref:`config_file@directory@project_directory` to be relative to the
:ref:`config_file@directory@rst_directory` .

file_name
*********
Is a file name relative to the project directory.
The entire command gets replaced by a name for the same file
relative to the rst directory.
Leading and trailing white space in *file_name* is ignored.

Example
*******
:ref:`dir_example-name`

{xrst_end dir_cmd}
"""
# ----------------------------------------------------------------------------
import re
import os
import xrst
#
# {xrst_begin dir_cmd_dev dev}
# {xrst_comment_ch #}
#
# Convert File Names to be Relative to the RST Directory
# ######################################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for this page.
#
# rst2project_dir
# ***************
# is a relative path from the :ref:`config_file@directory@rst_directory`
# to the :ref:`config_file@directory@project_directory` .
#
# data_out
# ********
# The return data_out is a copy of data_in except that all the occurrences of
# ``\{xrst_dir`` *file_name*\ ``}`` have been converted to the file name
# relative to the rst directory.
#
# {xrst_end dir_cmd_dev}
# BEGIN_DEF
def dir_command(data_in, rst2project_dir) :
   assert type(data_in) == str
   assert type(rst2project_dir) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_dir
   m_dir  = xrst.pattern['dir'].search(data_out)
   while m_dir != None :
      #
      # data_before, data_after
      data_before  = data_out[: m_dir.start()] + m_dir.group(1)
      data_after   = data_out[m_dir.end() :]
      #
      # data_left, data_out
      file_name    = m_dir.group(2).strip()
      file_name     = os.path.join(rst2project_dir, file_name)
      data_left    = data_before + file_name
      data_out     = data_left + data_after
      #
      # m_dir
      m_dir = xrst.pattern['dir'].search(data_out, len(data_left))
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN



#! /usr/bin/env python3
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-23 Bradley M. Bell
# -----------------------------------------------------------------------------
import re
import sys
import os
import toml
import tempfile
import subprocess
#
# pattern
pattern = dict()
pattern['index'] = re.compile( r'\n[.][.] index::[^\n]*' )
pattern['ref<>'] = re.compile( r':ref:`([^<`]*)<[^>`]*>`' )
pattern['ref@']  = re.compile( r':ref:`[^<`]*@([^<`@]*)`' )
pattern['ref-']  = re.compile( r':ref:`([^<@`]*)-(name|title)`' )
#
# main
def main() :
   #
   # rst_file, man_file
   if len(sys.argv) != 3 :
      msg = 'usage: bin/rst2man.py rst_file man_file'
      sys.exit(msg)
   rst_file = sys.argv[1]
   man_file = sys.argv[2]
   #
   if not os.path.isfile(rst_file) :
      msg = f'rst_file = {rst_file} is not a file'
      sys.exit(msg)
   #
   # rst_prolog
   if not os.path.isfile('xrst.toml') :
      msg = f'Cannot find xrst.toml in current directory'
      sys.exit(msg)
   file_obj    = open('xrst.toml', 'r')
   file_data   = file_obj.read()
   conf_dict   = toml.loads(file_data)
   rst_prolog  = conf_dict['include_all']['rst_prolog']
   file_obj.close()
   #
   # rst_data
   file_obj  = open(rst_file, 'r')
   rst_data  = file_obj.read()
   file_obj.close()
   #
   # temp_data
   temp_data = rst_prolog + rst_data
   temp_data = pattern['index'].sub(r'', temp_data)
   temp_data = pattern['ref<>'].sub(r'\1', temp_data)
   temp_data = pattern['ref@'].sub(r'\1', temp_data)
   temp_data = pattern['ref-'].sub(r'\1', temp_data)
   #
   # temp_file
   (fd, temp_file) = tempfile.mkstemp(
      suffix='.rst', prefix='rst2man_', dir='build', text=True
   )
   os.close(fd)
   file_obj  = open(temp_file, 'w')
   file_obj.write(temp_data)
   file_obj.close()
   #
   # man_file
   command = [ 'rst2man', temp_file, man_file ]
   subprocess.run(command)
   #
   # temp_file
   os.remove(temp_file)
#
main()



# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: Bradley M. Bell <bradbell@seanet.com>
# SPDX-FileContributor: 2020-24 Bradley M. Bell
# ----------------------------------------------------------------------------
r"""
{xrst_begin literal_cmd user}
{xrst_spell
   literalinclude
}

Literal Command
###############

Syntax
******

Entire File
===========
``\{xrst_literal}``

With Separator
==============
| ``\{xrst_literal`` *separator*
|     *display_file*
|     *start_after_1* *separator* *end_before_1*
|     *start_after_2* *separator* *end_before_2*
|     ...
| ``}``

Without Separator
=================
| ``\{xrst_literal``
|     *display_file*
|     *start_after_1*
|     *end_before_1*
|     *start_after_2*
|     ...
| ``}``

Purpose
*******
Literal text, from any where in any file,
can be included using this command.

literalinclude
**************
This command is similar to the following sphinx directive
(see :ref:`dir_cmd-name`) :

| |tab| .. literalinclude:: \{xrst_dir *display_file*}
| |tab| |tab| :start-after: *start_after_1*
| |tab| |tab| :end-before: *end_before_1*

The xrst literal command has the following difference:

#. If the *display_file* is not specified, the current
   :ref:`begin_cmd@Page File` is used.
#. If the *display_file* is the current :ref:`template_cmd@Input File` ,
   the *start_after* and *end_before* in the command are not considered
   a match for the corresponding text. This makes it possible to put a literal
   command in the same file as the text it will display.
#. It is an error for there to be more than one copy of each *start_after*
   or *end_before* in the *display_file* (not counting the copy in the
   command when the display file is the current input file).
   This makes sure that the intended section of *display_file* is displayed.
#. It is possible to specify multiple sections of a file using
   the start after and end before patterns.

Tokens
******
#. Leading and trailing spaces are not included in
   *separator*, *display_file*, each *start_after*, and each *end_before*.
#. Each *start_after* must have a corresponding *end_before*.
#. If there are an even number of tokens (not counting *separator*),
   the *display_file* is not present and the current page file is used.
#. The new line character separates the tokens.
#. If there are multiple lines in the command, the last line contains
   the ``}`` and must have nothing else but white space.


separator
*********
If *separator* is present, it must be a single character.
At most one *separator* can be in each line and it also separates tokens.

display_file
************
If *display_file* is not present,
the literal input block is in the current page file.
Otherwise, the literal input block is in *display_file*.
The file name *display_file* is relative to the
:ref:`config_file@directory@project_directory` .

1. This may seem verbose, but it makes it easier to write scripts
   that move files and automatically change references to them.
2. Note that if you use the sphinx ``literalinclude`` directive,
   the corresponding file name will be relative to the
   :ref:`config_file@directory@rst_directory` , which is a path relative
   to the project_directory; see :ref:`dir_cmd-name` .

extension
=========
The *display_file* extension is used to determine what language
to use when highlighting the input block.
In the special case where *display_file* ends with ``.in`` ,
the final ``.in`` is not included when file name
when determining the extension.
This is done because configure files use the ``.in`` extension,
and usually create a file with the ``.in`` dropped.

No start or end
***************
In the case where there is no *start_after* or *end_before*,
the entire display file is displayed.
In the case of the ``\{xrst_literal}`` syntax,
the entire current page file is displayed.

start_after
***********
Each literal input block starts with the line following the occurrence
of the text *start_after* in *display_file*.
If this is the same as the file containing the command,
the text *start_after* will not match any text in the command.
There must be one and only one occurrence of *start_after* in *display_file*,
not counting the command itself when the files are the same.

end_before
**********
Each literal input block ends with the line before the occurrence
of the text *end_before* in *display_file*.
If this is the same as the file containing the command,
the text *end_before* will not match any text in the command.
There must be one and only one occurrence of *end_before* in *display_file*,
not counting the command itself when the files are the same.

Spell Checking
**************
Spell checking is **not** done for these literal input blocks.


Example
*******
see :ref:`literal_example-name` .

{xrst_end literal_cmd}
"""
# ----------------------------------------------------------------------------
import os
import re
import xrst
#
# ----------------------------------------------------------------------------
#
# extension_map
# map cases that pygments has trouble with
extension_map = {
   'xrst' : 'rst'    ,
   'hpp'  : 'cpp'    ,
   'm'    : 'matlab' ,
   'txt'  : ''       ,
}
def file_extension(display_file) :
   if display_file.endswith('.in') :
      display_file = display_file[: -3]
      index        = display_file.rfind('.')
   else :
      index = display_file.rfind('.')
   extension = ''
   if 0 <= index and index + 1 < len(display_file) :
      extension = display_file[index + 1 :]
      if extension in extension_map :
         extension = extension_map[extension]
   return extension
#
# pattern_literal
pattern_literal    = xrst.pattern['literal']
#
# pattern_arg
pattern_arg        = re.compile( r'([^\n]*)@xrst_line ([0-9]+)@\n|\n' )
# ----------------------------------------------------------------------------
# {xrst_begin literal_cmd_dev dev}
# {xrst_comment_ch #}
#
# Process the literal commands in a page
# ######################################
#
# Prototype
# *********
# {xrst_literal ,
#    # BEGIN_DEF, # END_DEF
#    # BEGIN_RETURN, # END_RETURN
# }
#
# data_in
# *******
# is the data for a page before the
# :ref:`literal commands <literal_cmd-name>` have been removed.
#
# page_file
# *********
# is the name of the file that contains the begin command for this page.
# This is used for error reporting and for the display file
# when the display file is not included in the command and the command
# is not in a template expansion.
#
# page_name
# *********
# is the name of the page that this data is in. This is only used
# for error reporting.
#
# rst2project_dir
# ***************
# is a relative path from the :ref:`config_file@directory@rst_directory`
# to the :ref:`config_file@directory@project_directory` .
#
# data_out
# ********
# Each xrst literal command is converted to its corresponding sphinx commands.
#
# {xrst_end literal_cmd_dev}
# BEGIN_DEF
def literal_command(data_in, page_file, page_name, rst2project_dir) :
   assert type(data_in) == str
   assert type(page_file) == str
   assert type(page_name) == str
   assert type(rst2project_dir) == str
   # END_DEF
   #
   # data_out
   data_out = data_in
   #
   # m_literal
   m_literal  = pattern_literal.search(data_out)
   while m_literal != None :
      #
      # separator
      if m_literal.group(1) == None :
         separator = ''
      else :
         separator = m_literal.group(1).strip()
      if len(separator) > 1 :
         msg  =  '{xrst_literal separator\n'
         msg += f'separator = "{separator}" is more than one character'
         xrst.system_exit(
            msg,
            file_name = page_file,
            page_name = page_name,
            m_obj     = m_literal,
            data      = data_out
         )
      #
      # arg_list, m_list
      arg_list  = list()
      m_list    = list()
      if m_literal.group(2) != None :
         #
         # m_arg
         index = data_out.find('\n', m_literal.start() + 1)
         m_arg = pattern_arg.search(data_out, index + 1)
         while m_arg != None and m_arg.end() < m_literal.end() :
            if m_arg.group(1) == None :
               msg = 'There is an empty line inside a literal command.'
               xrst.system_exit(
                  msg,
                  file_name = page_file,
                  page_name = page_name,
                  m_obj     = m_arg,
                  data      = data_out
               )
            #
            # arg, line
            arg    = m_arg.group(1)
            if len(separator) > 0 and arg.count(separator) > 1 :
               msg  =  f'xrst_literal separator = {separator}\n'
               msg += 'separator appears more than once in a line'
               xrst.system_exit(
                  msg,
                  file_name = page_file,
                  page_name = page_name,
                  m_obj     = m_arg,
                  data      = data_out,
               )
            #
            # arg_list, m_list
            if separator == '' or arg.count(separator) == 0 :
               arg_list.append( arg.strip() )
               m_list.append(m_arg)
            else :
               assert arg.count(separator) == 1
               arg = arg.split(separator)
               #
               arg_list.append( arg[0].strip() )
               m_list.append(m_arg)
               #
               arg_list.append( arg[1].strip() )
               m_list.append(m_arg)
            #
            # m_arg
            m_arg  = pattern_arg.search(data_out , m_arg.end() )
      #
      # cmd_line
      if len(arg_list) >= 2 :
         start_line = int( m_list[0].group(2) )
         end_line  = int( m_list[-1].group(2) )
         cmd_line = ( start_line, end_line )
      #
      # input_file
      page_line, template_file, template_line = \
         xrst.file_line(m_literal, data_out)
      if template_file == None :
         input_file = page_file
      else :
         input_file = template_file
      #
      # even
      even = len(arg_list) % 2 == 0
      #
      # display_file, arg_list
      if even :
         display_file = page_file
      else :
         display_file = arg_list.pop(0)
         m_arg        = m_list.pop(0)
         if not os.path.isfile(display_file) :
            msg  = 'literal command: can not find the display_file.\n'
            msg += f'display_file = {display_file}'
            xrst.system_exit(msg,
               file_name = page_file,
               page_name = page_name,
               m_obj     = m_arg,
               data      = data_out
            )
         if os.path.samefile(display_file, page_file) :
            display_file = page_file
      #
      # start_end_line_list
      assert len(arg_list) % 2 == 0
      start_end_line_list = list()
      for i in range(0, len(arg_list), 2) :
         start_after = arg_list[i]
         end_before  = arg_list[i+1]
         #
         # start_line, end_line
         start_line, end_line = xrst.start_end_file(
            page_file    = page_file,
            page_name    = page_name,
            input_file   = input_file,
            display_file = display_file,
            cmd_line     = cmd_line,
            start_after  = start_after,
            end_before   = end_before,
            m_start      = m_list[i],
            m_end        = m_list[i+1],
            m_data       = data_out,
         )
         #
         # start_end_line_list
         start_end_line_list.append( (start_line + 1, end_line - 1) )
      #
      # cmd
      display_path = os.path.join(rst2project_dir, display_file)
      cmd          = f'.. literalinclude:: {display_path}\n'
      #
      # cmd
      for i in range( len(start_end_line_list) ) :
         start_line, end_line = start_end_line_list[i]
         if i == 0 :
            cmd += 2 * ' ' + f':lines: {start_line}-{end_line}'
         else :
            cmd += f',{start_line}-{end_line}'
      if( len(start_end_line_list) > 0 ) :
         cmd += '\n'
      #
      # cmd
      # Add language to literalinclude, sphinx seems to be brain
      # dead and does not do this automatically.
      extension = file_extension( display_file )
      if extension != '' :
         cmd += 2 * ' ' + f':language: {extension}\n'
      cmd = '\n' + cmd + '\n\n'
      if m_literal.start() > 0 :
         if data_out[m_literal.start() - 1] != '\n' :
            cmd = '\n' + cmd
      #
      # data_tmp, data_out
      data_tmp  = data_out[: m_literal.start() ]
      data_tmp += cmd
      data_out  = data_tmp + data_out[ m_literal.end() : ]
      #
      # m_literal
      m_literal  = pattern_literal.search(data_out, len(data_tmp))
   #
   # BEGIN_RETURN
   #
   assert type(data_out) == str
   return data_out
   # END_RETURN
